/* Everything to maintain the field, which is a wrapping matrix
   containing all mass pixels. It's a kind of map, and usually much
   bigger than the visible part of the screen. */
#ifndef __FIELD__
#define __FIELD__

/* FIXME: is this really needed? It's only for the struct mass, which
   is only referenced by pointer anyway... */
#include "mass.h"

/* Needs to be a power of 2 */
#define FIELDSIZE 1024
/* So 'x & FIELDWRAP' will wrap around the border (also for x<0).
   See also declaration of 'field' below for more comments. */
#define FIELDWRAP (FIELDSIZE-1)

/* the field consists of a frozen area, never visible, where masses
   are just being around to thaw when the player approaches */
#define FIELDFREEZE 512


/* This ist *the* field, where the game takes place.  It's an array
 * with wrapping borders, dimensions FIELDSIZE*FIELDSIZE. The wrapping
 * is handled automagically by the FieldXY() and FieldPoint() macros
 * below, but it's somewhat more tricky for comparing coordinates and
 * measuring distances.
 *
 * I've chosen a periodical field in order to make the field (from the
 * player's point of view) unlimited. The idea is to leave a big part
 * of the field frozen, and update only the regions around the
 * player. While he moves, either old masses will be thawn, or new
 * onces generated just in time. (PLANNED: a 'big' map, so when the
 * player moves up and then down, it will at least look similar).
 */
typedef struct _FieldPoint_ {
  int height;
  Mass * owner;
} FieldPoint;
extern FieldPoint * field;

/* the point of interest, most likely the player's position;
   this defines the center of the thawn field */
extern Point fieldCenter;
/* TODO: this will be the center point of all *local* players, so an
   internet player can get out of the screen without disturbing local
   players. This must not influence the engine, and might scroll all
   around while debugging, so that's the wrong place for it. */
#define displayCenter fieldCenter
/* direction where the player looks at */
extern int displayHeading;

/* accessing the field */
#define FieldXY(x, y) (field + FIELDSIZE*((y)&(FIELDWRAP)) + ((x)&(FIELDWRAP)))
/* FIXME: name conflict! FieldPoint is already used. Strange this works. */
#define FieldPoint(p) FieldXY(p.x, p.y)
extern Point FieldToPoint(FieldPoint * p);

/* for main() */
extern void InitFieldMemory();
extern void FreeFieldMemory();
extern void CleanField();
extern void InitMassesMemory();
extern void FreeMassesMemory();

/* returns a NULL terminated list of pointers to masses that at least
   touch the rect r with their maxrect; the list must bee freed */
extern Mass ** GetMassesInRect(Rect r);
/* returns the rect in which masses are not frozen */
extern Rect GetThawnRect();

/* check wether the pointer is valid; for debugging only, it is slow */
extern int IsValidMass(Mass * m);
/* Expensive! */
extern void DebugCheckField();

/* some math2d.h functions adapted to wrapping in the field */

extern Rect Wrap_SmallestCommonRect(Rect r1, Rect r2);
/* move p to the nearest border of r, if it isn't already part of r */
extern void Wrap_FitPointIntoRect(Point * p, Rect r);
extern int Wrap_IsPointInRect(Point p, Rect r);

extern int Wrap_GetCompareDistance(Point p1, Point p2);
extern double Wrap_GetRealDistance(Point p1, Point p2);

extern void Wrap_ExpandRectToIncludePoint(Rect * r, int x, int y);
extern Point Wrap_VectorFromTo(Point from, Point to);

/* IMPORTANT NOTE: considering network games, randarea() MUST be used
   for all in-game random decisions which have to go synchronous on all
   clients; do not use in display code! */
extern int randarea(int smallest, int size);

#endif
