#include <stdlib.h>
#include <stdio.h>
#include "game.h"
#include "masses.h"
#include "mass.h"
#include "field.h"
#include "networkplayer.h"
#include "debug.h"

int frame = 0;			/* frames past since start */
int paused = 0;
//int singleFrame = 0;
int exitFrame = 0;		/* automatically exit then */

/* prototypes */
void CreateBombermanGame();
void CreateSwarmGame();
void CreateDebugGame();
void CreateTestGame();
void CreateProfilingGame();

void ShowMenu()
{
  /* Will there ever be a Menu?
     Or will the menu be a seperate program, calling this one? */
}

int StartGame(int argc, char ** argv)
{
  int i;
  int gameCreated = 0;
  for (i=0; i<argc; i++) {
    if (!strcmp(argv[i], "--game")) {
      gameCreated = 1;
      i++;
      if (i == argc) {
	printf("Try %s --game list\n", argv[0]);
	exit(1);
      }
      if (!strcmp(argv[i], "list") || !strcmp(argv[i], "help")) {
	printf("Usage: --game <gametype> 

Where <gametype> is one of:

swarm        control the swarm
bomberman    walls in bomberman style
testgame     default, does whatever I'm working on
profile      speed measurement, no interaction possible
");
	exit(1);
      } else if (!strcmp(argv[i], "bomberman")) {
	CreateBombermanGame();
      } else if (!strcmp(argv[i], "swarm")) {
	CreateSwarmGame();
      } else if (!strcmp(argv[i], "profile")) {
	CreateProfilingGame();
      } else if (!strcmp(argv[i], "testgame")) {
	CreateTestGame();
      } else {
	printf("Unknown game type: %s\n", argv[i]);
	exit(1);
      }
    }
  }

  if (!gameCreated) CreateTestGame();
  return 0;
}

void CreateDebugGame()
{
}

void CreateBombermanGame()
{
  int x,y;
  for (x=0; x<10; x++) for (y=0; y<10; y++) {
    int px, py;
    /* fit the game onto the screen */
#define SCREENSMALL FIELDSIZE
    px = SCREENSMALL*(x+2)/14;
    py = SCREENSMALL*(y+2)/14;
    if (x == 2 && y == 5) {
      CreatePlayerMass(px, py, 0);
      /* TODO: do not add players here. Add them on keypress. */
      AddLocalPlayer(0);
    } else if (x == 7 && y == 5) {
      CreatePlayerMass(px, py, 1);
    } else if (randarea(0, 3) == 0) {
      CreateLivingWallMass(px, py);
    } else if (randarea(0, 6) == 0) {
      CreateGrowingFoodMass(px, py);
    }
  }
}

void CreateSwarmGame()
{
  int i, x, y;
  Mass * m;
  /* 1. create some walls */
  for (i=0; i<20; i++) {
    x = randarea(10, FIELDSIZE/2-20);
    y = randarea(10, FIELDSIZE/2-20);
    m = CreateLivingWallMass(x, y);
    //m->speed = randarea(10, 50) / 256.0;
  }
  /* 2. create some food */
  for (i=0; i<5; i++) {
    x = randarea(10, FIELDSIZE/2-20);
    y = randarea(10, FIELDSIZE/2-20);
    m = CreateMassNearby(x, y, 3);
    m->playerFood = 1;
    m->speed = 222;
    m->growRounds = 130;
    RGBSET(m->baseColor, 120, 255, 100);
  }
  /* 3. create the player */
  AddLocalPlayer(0);
  CreatePlayerMass(100, 100, 0);
  /* 4. create the swarm */
  for (i=0; i<20; i++) {
    x = randarea(50, 40);
    y = randarea(50, 40);
    m = CreateCellMass(x, y);
  }
}

void CreateTestGame()
{
  int i, x, y;
  //Mass * m;

  //CreateSwarmGame();
  //CreatePlayerMass(10, 10, 0);
  CreatePlayerMass(200, 200, 0);
  AddLocalPlayer(0);
  CreateNexusMass(240, 240);
  return;

  for (x=0; x<15; x++) for (y=0; y<15; y++) {
    //int px, py;
#define SCREENSMALL FIELDSIZE
    /*
    px = SCREENSMALL*(x+2)/24;
    py = SCREENSMALL*(y+2)/24;
    m = CreateCellMass(px, py);
    */
    /*
    m->baseLineGrowth->height = (x-5)*HEIGHT_NORM/10;
    m->baseLineGrowth->start = 
      m->baseLineGrowth->end = (y-5)*HEIGHT_NORM/6;
    m->growRounds /= 2;
    */
    /*
    m->baseLineGrowth = calloc(1, sizeof(BaseLineGrowth));
    m->baseLineGrowth->step = 128;
    m->baseLineGrowth->height = 0;
    m->baseLineGrowth->start = HEIGHT_NORM/4;
    m->baseLineGrowth->end = 0;
    */
  }

  //CreatePlant2Mass(50, 55);

  for (i=0; i<90; i++) {
    x = randarea(10, FIELDSIZE/3-20);
    y = randarea(10, FIELDSIZE/3-20);
    CreateCellMass(x, y);
  }

  //CreateAvoidPointMass(30, 30);
  //CreateCoinMass(50, 80);
  AddLocalPlayer(0);
  CreatePlayerMass(100, 100, 0);
  //CreateLivingWallMass(80, 120);
  /*
  for (i=0; i<4; i++) {
    x = randarea(100, SCREENX-200);
    y = randarea(90, SCREENY-180);
    CreateMazeWall(x, y);
  }
  MazeGenerator();
  */
}

void CreateProfilingGame()
{
  int i, x, y;
  srand(0);
  exitFrame = 1000;
  for (i=0; i<40; i++) {
    x = randarea(10, FIELDFREEZE-20);
    y = randarea(10, FIELDFREEZE-20);
    CreateGrowingFoodMass(x, y);
  }
  for (i=0; i<40; i++) {
    x = randarea(10, FIELDFREEZE-20);
    y = randarea(10, FIELDFREEZE-20);
    CreateLivingWallMass(x, y);
  }
  for (i=0; i<30; i++) {
    CreateAvoidPointMass(x, y);
  }
  /* since no player there... */
  fieldCenter.x = FIELDFREEZE/2;
  fieldCenter.y = FIELDFREEZE/2;
}

int NextGameFrame()
{
  frame++;
  ContinueMasses();
  if (frame == exitFrame) {
    printf("Frame %d, game checksum is %d.\n", frame, FieldChecksum());
    return 1;
  }
  return 0;
}

