#include "input.h"
#include "debug.h"
#include "SDL.h"
/* all these things just for mkdir()... */
/* TODO: use SDL's file handling */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

/* globals */
KeyStates key;
ConfigData config;

/* SCRIPT_START_KEYBOARD_OUTPUT2 */
#define NUM_KEYS 21
/* SCRIPT_END */

/* implementation */
typedef struct {
  char * name;			/* symbolic name */
  char * def;			/* default */
  char * desc;			/* long description */
} KeyInfo;
KeyInfo keyInfo[NUM_KEYS] = {
  /* SCRIPT_START_KEYBOARD_OUTPUT3 */
  {"quit", "escape", "quit the game"},
  {"player1left", "left", "player 1 left"},
  {"player1up", "up", "player 1 up"},
  {"player1right", "right", "player 1 right"},
  {"player1down", "down", "player 1 down"},
  {"player1button1", "space", "player 1 button 1"},
  {"player1button2", "n", "player 1 button 2"},
  {"player2left", "a", "player 2 left"},
  {"player2up", "w", "player 2 up"},
  {"player2right", "d", "player 2 right"},
  {"player2down", "s", "player 2 down"},
  {"player2button1", "tab", "player 2 button 1"},
  {"player2button2", "q", "player 2 button 2"},
  {"pause", "pause", "pause the game"},
  {"nextFrame", "return", "when paused, step by one frame"},
  {"switchInterface", "f10", "switch between 2D and 3D graphics"},
  {"debugHUD", "f2", "show debug head up display (DHUD)"},
  {"debugHUDzin", "[+]", "DHUD zoom in"},
  {"debugHUDzout", "[-]", "DHUD zoom out"},
  {"debugHUDcycle", "f3", "DHUD cycle display mode"},
  {"debugHUDrect", "f4", "DHUD show maxRect"},
  /* SCRIPT_END */
};

/* one physical key can trigger serveral "virtual" game keys at once */
KeyState ** keyMap[SDLK_LAST];

MouseState mouse;

/* system dependent path to append the config file name to */
char confpath[200];

void LoadConfig();
void InitInput() {
  /* Pathing stuff for the different operating systems */
#if defined(unix) || defined(__MACH__)
  strcpy(confpath, "");
  strcat(confpath, getenv("HOME"));
  strcat(confpath, "/.mass/");  
#elif defined(WIN32)
  strcpy(confpath, "");
#elif defined(macintosh)
  /* something with : , but I really don't know */
#error Do you want to port it to Mac?
#else
#error Unspecified platform
#endif /* Choose your platform */
  SDL_EnableKeyRepeat(0, 0);
  memset(&mouse, 0, sizeof(mouse));
  memset(keyMap, 0, sizeof(keyMap));
  ResetKeys();
  LoadConfig();
}

void SaveConfig();
void DoneInput() {
  SaveConfig();
}

SDLKey missingInSDL_GetKeyByName(char * s)
{
  SDLKey i;
  for (i=0; i<SDLK_LAST; i++) {
    if (strcmp(s, SDL_GetKeyName(i)) == 0) return i;
  }
  return SDLK_UNKNOWN;
}

void ListValidSDLKeys()
{
  int i;
  printf("Valid key names are: ");
  for (i=0; i<SDLK_LAST; i++) {
    if (strcmp("unknown key", SDL_GetKeyName(i))) {
      printf("'%s' ", SDL_GetKeyName(i));
    }
  }
  printf("\n");
}

SDLKey GetSDLKeyOf(int gameKey) {
  /* well, since this information is not needed while playing, its
     somewhat hidden ;-) */
  SDLKey i;
  for (i=0; i<SDLK_LAST; i++) {
    if (keyMap[i]) {
      int j;
      for (j=0; keyMap[i][j]; j++) {
	if (keyMap[i][j] == (KeyState*)(&key)+gameKey) return i;
      }
    }
  }
  return SDLK_UNKNOWN;
}

void SetSDLKeyOf(int gameKey, SDLKey realKey) {
  /* 1. remove old keybinding */
  SDLKey i;
  for (i=0; i<SDLK_LAST; i++) {
    if (keyMap[i]) {
      int j, pos;
      pos = -1;
      for (j=0; keyMap[i][j]; j++) {
	if (keyMap[i][j] == (KeyState*)(&key)+gameKey) pos = j;
      }
      if (pos != -1) {
	if (j==2) {
	  /* no more gamekey bound to this one */
	  free(keyMap[i]);
	  keyMap[i] = NULL;
	} else {
	  keyMap[pos] = keyMap[j-1];
	  keyMap[j-1] = NULL;
	}
      }
    }
  }
  /* 2. add new keybinding */
  if (!keyMap[realKey]) {
    keyMap[realKey] = calloc(2, sizeof(KeyState*));
    keyMap[realKey][0] = (KeyState*)(&key)+gameKey;
  } else {
    int j;
    for (j=0; keyMap[i][j]; j++) ;
    keyMap[realKey] = realloc(keyMap[i], (j+2)*sizeof(KeyState*));
    keyMap[realKey][j] = (KeyState*)(&key)+gameKey;
    keyMap[realKey][j+1] = NULL;
  }
}

void SaveConfig() {
  FILE * f;
  char filename[250];
  int i;
  char * s;
  /* create config directory */
#if defined(unix) || defined(__MACH__)
  mkdir(confpath, 0777); 
#else
  mkdir(confpath); 
#endif

  strcpy(filename, confpath);
  strcat(filename, "keyboard");
  f = fopen(filename, "w+");
  if (!f) {
    DebugLevel(0, "Could not write config file '%s'.\n", filename);
  } else {
    for (i=0; i<NUM_KEYS; i++) {
      s = keyInfo[i].desc;
      fwrite("# ", 1, 2, f);
      fwrite(s, strlen(s), 1, f);
      fwrite("\n", strlen("\n"), 1, f);

      s = keyInfo[i].name;
      fwrite(s, strlen(s), 1, f);
      fwrite("=", 1, 1, f);
      /* look it up */
      s = SDL_GetKeyName(GetSDLKeyOf(i));
      fwrite(s, strlen(s), 1, f);
      fwrite("\n", strlen("\n"), 1, f);
    }
    fclose(f);
  }
}

void LoadConfig() {
  FILE * f;
  char filename[250];
  int i;
  int badkey=0;

  /* reset all keybindings to default */
  for (i=0; i<NUM_KEYS; i++) {
    SDLKey k;
    k = missingInSDL_GetKeyByName(keyInfo[i].def);
    if (k!=SDLK_UNKNOWN) {
      SetSDLKeyOf(i, k);
    } else {
      DebugLevel(0, "Unknown default SDL Key for '%s': '%s'.\n", keyInfo[i].name, keyInfo[i].def);
      badkey=1;
    }
  }
  if (badkey) ListValidSDLKeys();

  strcpy(filename, confpath);
  strcat(filename, "keyboard");
  f = fopen(filename, "r");
  if (!f) {
    DebugLevel(1, "Could not read config file '%s', using defaults.\n", filename);
  } else {
    SDLKey k;
    int linenr=0;
    char line[250];
    char * p;
    char * value;
    int found_i;
    while (fgets(line, 250, f)) {
      linenr++;
      /* strip \n and windows newlines */
      if ((p = strchr(line, '\n'))) *p = '\0';
      if ((p = strchr(line, '\r'))) *p = '\0';

      value = strchr(line, '=');
      if (strlen(line) > 2 && value && line[0] != '#') {
	*(value++) = '\0'; /* now the 'line' string is terminated where the first space was */
	if (strlen(line) > 0 && strlen(value) > 0 ) {
	  found_i = 0;
	  for (i=0; i<NUM_KEYS; i++) {
	    if (strcmp(line, keyInfo[i].name) == 0) {
	      found_i++;
	      k = missingInSDL_GetKeyByName(value);
	      if (k != SDLK_UNKNOWN) {
                SetSDLKeyOf(i, k);
	      } else {
		DebugLevel(0, "%s: unknown key description '%s' in line %d\n", filename, value, linenr);
	      }
	    }
	  }
	  if (found_i == 0) {
	    /* TODO: check for config strings other than keys */
	    DebugLevel(0, "%s: unknown name '%s' in line %d ignored.\n", filename, line, linenr);
	  }
	} else {
	  DebugLevel(2, "%s: skiped strange short line %d\n", filename, linenr);
	}
      } else {
	DebugLevel(2, "%s: skiped line %d\n", filename, linenr);
      }
    }
    fclose(f);
  }
  /* key bindings have changed, better reset the key states */
  ResetKeys();
}

void ResetKeys() {
  int i;
  for (i=0; i<NUM_KEYS; i++) {
    memset((KeyState*)(&key)+i, 0, sizeof(KeyState));
  }
}

void KeyStateToggle(KeyState * p, int newState)
{
  if (newState == p->pressed) return;
  p->pressed = !p->pressed;
  if (p->pressed) {
    p->hit = 1;
    p->hits++;
    p->toggled = !p->toggled;
  }
}

void UpdateInput() {
  int i;
  SDL_Event event;
  KeyState * p;

  for (i=0; i<NUM_KEYS; i++) ((KeyState*)(&key)+i)->hit = 0; /* OPTIMIZE perhaps? */
  mouse.left.hit = mouse.right.hit = mouse.middle.hit = 0;

  // process events
  while ( SDL_PollEvent(&event) ) {
    p = NULL;
    switch (event.type) {
    case SDL_QUIT:
      /* emulate hit of the quit key */
      KeyStateToggle(&key.quit, 1);
      break;
    case SDL_KEYDOWN:
    case SDL_KEYUP:
      if (keyMap[event.key.keysym.sym]) {
	for (i=0; (p=keyMap[event.key.keysym.sym][i]); i++) {
	  KeyStateToggle(p, (event.type==SDL_KEYDOWN)?1:0);
	}
      }
      break;
    case SDL_MOUSEBUTTONDOWN:
    case SDL_MOUSEBUTTONUP:
      if (event.button.button == SDL_BUTTON_RIGHT) p = &mouse.right;
      if (event.button.button == SDL_BUTTON_LEFT) p = &mouse.left;
      if (event.button.button == SDL_BUTTON_MIDDLE) p = &mouse.middle;
      if (p) KeyStateToggle(p, (event.type==SDL_MOUSEBUTTONDOWN)?1:0);
      break;
    case SDL_MOUSEMOTION:
      mouse.pos.x = event.motion.x;
      mouse.pos.y = event.motion.y;
      break;
    default:
      break;
    }
  }
}
