#ifndef __INPUT_H__
#define __INPUT_H__

#include "SDL.h"
#include "math2d.h"

/* Add new keys here, then run the script newkeys.py. Format:
name             default    description
SCRIPT_START_KEYBOARD_INPUT
quit             "escape"   "quit the game"
player1left      "left"     "player 1 left"
player1up        "up"       "player 1 up"
player1right     "right"    "player 1 right"
player1down      "down"     "player 1 down"
player1button1   "space"    "player 1 button 1"
player1button2   "n"        "player 1 button 2"
player2left      "a"        "player 2 left"
player2up        "w"        "player 2 up"
player2right     "d"        "player 2 right"
player2down      "s"        "player 2 down"
player2button1   "tab"      "player 2 button 1"
player2button2   "q"        "player 2 button 2"
pause            "pause"    "pause the game"
nextFrame        "return"   "when paused, step by one frame"
switchInterface  "f10"      "switch between 2D and 3D graphics"
debugHUD         "f2"       "show debug head up display (DHUD)"
debugHUDzin      "[+]"      "DHUD zoom in"
debugHUDzout     "[-]"      "DHUD zoom out"
debugHUDcycle    "f3"       "DHUD cycle display mode"
debugHUDrect     "f4"       "DHUD show maxRect"
SCRIPT_END
*/

typedef struct {
  /* those are read/write from everywhere */
  unsigned int hit     : 1;  /* key is pressed down and wasn't before */
  unsigned int pressed : 1;  /* key is (still) hold down now */
  unsigned int toggled : 1;  /* gets inverted when hit */
  unsigned int hits;         /* total keystrokes since start of the game */
} KeyState;

typedef struct  {
  /* DO NOT EDIT - see above. */
  /* SCRIPT_START_KEYBOARD_OUTPUT1 */
  KeyState quit, player1left, player1up, player1right, player1down, player1button1, player1button2, player2left, player2up, player2right, player2down, player2button1, player2button2, pause, nextFrame, switchInterface, debugHUD, debugHUDzin, debugHUDzout, debugHUDcycle, debugHUDrect;
  /* SCRIPT_END */
} KeyStates;
extern KeyStates key;

/* Mouse */
typedef struct {
  KeyState left;
  KeyState middle;
  KeyState right;
  Point pos;
} MouseState;
extern MouseState mouse;


typedef struct {
  /* all configurations that are no keys */
  /* expected config file look:
     meta-server=
     autoconnect-to=10.0.0.1
     minimum-speed=0.5
     autostart-game=bomberman
     opengl-start=yes
     opengl-perspective=first
   */
  /* ...similar script as with the keyboard:
name          default
autostartGame "bomberman"
autostartGame "bomberman"
  */
} ConfigData;
extern ConfigData config;

extern void InitInput();
extern void UpdateInput();
//extern void SaveConfig();
extern void DoneInput();

/* all keys are set to untoggled, never hit, etc. */
extern void ResetKeys();

#endif

