/* TODO: *useful* Documentation
 * - overview (what is done, how connected)
 * - goals, things to do
 *  
 * This is the "core" file - which means it's my main playground,
 * messy, and does much too much.
 *
 *
 *
 */

#include "mass.h"
#include "field.h"
#include "masses.h"
#include "debug.h"
#include "networkplayer.h"
#include "math2d.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

// global parameters
#define PLAYER_MOVE_NIBBLES

Mass * CreateMassNearby(int x, int y, int radius)
{
  FieldPoint * p;
  Mass * m;
  /* search radius region */
  int i;
  Point pt;
  for (i=50; i>0; i--) {
    pt.x = randarea(x-radius, 2*radius);
    pt.y = randarea(y-radius, 2*radius);
    p = FieldPoint(pt);
    if (!p->owner) break;
  }
  if (i == 0) return NULL;
  x = pt.x;
  y = pt.y;

  assert(FieldXY(x, y) == p);
  m = CreateMass();
  p->owner = m;
  p->height = HEIGHT_NORM*8; /* so the pixel won't just dissappear */
  m->numPixels = 1;
  m->maxRect.x = x;
  m->maxRect.y = y;
  m->maxRect.h = 1;
  m->maxRect.w = 1;
  return m;
}

int ContinueMass(Mass * m);
int ReThink(Mass * m);
int PreFlattening(Mass * m);
void DoFlattening(Mass * m);

void ScreenBorderGuard();

void ContinueMasses()
{
  Mass * m;
  Mass ** massList;
  int i;
  /* TODO: handle death and birth during these routines */
  
  /* TODO (idea): perhaps process the list once forward and once backward,
     so there is more justice. (for the points of conflict) */

  massList = GetMassesInRect(GetThawnRect());
  for (i=0; (m=massList[i]); i++) {
    int dead;
    
    /* note: speed == 0 could make sense, but it's not yet used. */
    assert(m->speed > 0);
    assert(m->speed <= 1000*256);
    assert(m->pos < 256 && m->pos >= 0);

    m->pos += m->speed;

    /* TODO: in case there are masses with big speed values it could
       be better to process them in turn, to allow more
       interactions. */
    dead = 0;
    while (!dead && m->pos >= 256) {
      m->pos -= 256;
      dead = ReThink(m);
      if (!dead) DebugCheckMass(m);
      if (!dead) dead = ContinueMass(m);
      if (!dead) m->mustRedraw = 1;
    }
  }
  free(massList);
}

void RenewInterfaceReactions(Mass * m)
{
  /* add code that destroys/creates masses to ReThink(), not here */
  MassInterface * mi;
  mi = m->interfaces;
  while (mi) {
    Mass * m2 = mi->m;
    if (mi->m) {
      if (m->isPlant) {
	if (m2->isPlant) {
	  mi->borderReaction = 256/2;
	}
	if (m2->player != -1) {
	  mi->borderReaction = -1000;
	}
      }
      if (m->playerFood) {
	if (m2->player != -1) {
	  mi->borderReaction = 500;
	  mi->eatme = 1;
	}
      }
      if (m->nexusLevel && m2->nexusLevel) {
	mi->borderReaction = 256 - MIN(MIN(m->age, m2->age), 2*256);
	if (m->nexusLevel == 4 && m2->nexusLevel == 5 && m->age > 300 && mi->contact > 0) {
	  m->nexusLevel = 3;
	}
	if (m->nexusLevel == 3 && m2->nexusLevel == 5) mi->borderReaction = 300;
	if (m->nexusLevel == 5 && m2->nexusLevel == 3) mi->borderReaction = 300;
      }
      if (!mi->age++) {
	/* things which need to be done only once (or only when mi->age
	   is reset) */
      }
    }
    mi = mi->next;
  }

}

int ReThink(Mass * m)
{
  /* handle growth (temporary) */
  /* and avoid spikes (temporary too) */
  /*
  if (m->numPixels && m->growRounds && (m->volume / m->numPixels < HEIGHT_NORM*20)) {
    m->growRounds--;
    m->bufferMass += HEIGHT_NORM*3/2;
  }
  */
  if (m->growRounds) {
    m->growRounds--;
    m->bufferMass += HEIGHT_NORM*3/2;
    if (m->numPixels < 4) {
      /* boost at the start, because flattening of one lonely pixel
	 will shrink it to 1/9 of its height immediately */
      m->bufferMass += 9*HEIGHT_NORM;
    } 
    if (m->superGrow) {
      m->bufferMass += m->numPixels * HEIGHT_NORM/2;
    }
  }
  return 0;
}

void DebugPrintMass(Mass * m)
{
  int x, y;
  DebugLevel(3, "Mass %X %s:\n", m, RectString(m->maxRect));
  for (y=m->maxRect.y-1; y<m->maxRect.y+m->maxRect.h+2; y++) {
    for (x=m->maxRect.x-1; x<m->maxRect.x+m->maxRect.w+2; x++) {
      FieldPoint * p = FieldXY(x, y);
      if (p->owner == m) {
	DebugLevel(3, "M");
      } else if (p->owner == NULL) {
	DebugLevel(3, "-");
      } else {
	DebugLevel(3, "?");
      }
    }
    DebugLevel(3, "\n");
  }
}

void DebugCheckMass(Mass * m)
{
#ifdef DEBUG
  int x, y;
  int empty;
  int n;
  assert(m);
  empty = 0;
  if (m->numPixels == 0) empty++;
  if (m->maxRect.w <= 0 || m->maxRect.h <= 0) empty++;
  assert(m->maxRect.w >= 0);
  assert(m->maxRect.h >= 0);
  assert(m->maxRect.w <= FIELDSIZE);
  assert(m->maxRect.h <= FIELDSIZE);
  if (empty == 2) {
    DebugLevel(1, "Warning: empty mass %X still being around\n", m);
    return;
  }
  if (empty) {
    if (m->numPixels == 0) {
      DebugLevel(1, "Warning: empty mass %X has still with*height=%d. Fixed.\n", m, m->maxRect.w * m->maxRect.h);
      m->maxRect.w = 0;
      m->maxRect.h = 0;
    } else {
      DebugLevel(1, "Warning: empty mass %X has still %d numPixels\n", m, m->numPixels);
    }
    return;
  }
  assert(m->numPixels >= 0);
  n = 0;
  for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
    for (x=m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
      FieldPoint * p = FieldXY(x, y);
      if (p->owner == m) {
	assert(p->height >= 0);
	n++;
      }
    }
  }
  assert(m->numPixels == n);
#endif
}

/* parts belonging to PreFlattening: */
//void PreFlattening_Movement(Mass * m);
void PreFlattening_ShapePoints(Mass * m)
{
  int x, y;
  int i;
  FieldPoint * p;
  /* TODO: evaluate other algos */
  /* this one does rectangular shapepoints */
  for (i=0; i<MAX_SHAPEPOINTS; i++) {
    ShapePoint sp = m->shapePoints[i];
    if (sp.radius > 0 && sp.strength != 0) {
      Rect r;
      int radiusReciproc;
      radiusReciproc = sp.strength*256/sp.radius;
      r.x = sp.pos.x - sp.radius;
      r.y = sp.pos.y - sp.radius;
      r.w = r.h = sp.radius*2+1;
      r = Wrap_SmallestCommonRect(m->maxRect, r);
      for (y=r.y; y<r.y+r.h; y++) {
	for (x=r.x; x<r.x+r.w; x++) {
	  p = FieldXY(x, y);
	  if (p->owner == m) {
	    // TODO: wrap distance.
	    int distance = abs(x-sp.pos.x) + abs(y-sp.pos.y);
	    /* OPTIMIZE: redundant check: */
	    if (distance < sp.radius) {	
	      int addHeight;
	      addHeight = (sp.radius - distance) * radiusReciproc / 256;
	      p->height += addHeight;
	      m->bufferMass -= addHeight;
	      if (p->height < 0) {
		/* OPTIMIZE: remove multiple such checks and do them once
		   before (or even after?) the flattening. */
		m->bufferMass += p->height;
		p->height = 0;
	      }
	    }
	  }
	}
      }
    }
  }
}

/* TODO: move the functions below to a better location and announce
   them in an header file */

int CountShapePoints(Mass * m)
{
  int i, num=0;
  for (i=0; i<MAX_SHAPEPOINTS; i++)
    if (m->shapePoints[i].radius != 0 && m->shapePoints[i].strength != 0) num++;
  return num;
}
/* condition: -1 only negative strength, 0 everything, +1 only positive strength
   max: 1 search maximum, 0 search minimum */
int GetStrongestShapePoint_Backend(Mass * m, int condition, int max)
{
  int best = -1;
  int best_value = -1;
  ShapePoint sp;
  int i;
  for (i=0; i<MAX_SHAPEPOINTS; i++) {
    sp = m->shapePoints[i];
    if (sp.radius >= 0 && sp.strength != 0) {
      int value = sp.strength*sp.radius;
      if (condition == 0) value = abs(value); else value *= condition;
      if (value > 0) {
	/* condition is met */
	if (!max) value = -value;
	if (best == -1 || value > best_value) {
	  best_value = value;
	  best = i;
	}
      }
    }
  }  
  return best;
}
int GetStrongestShapePoint(Mass * m)
{
  return GetStrongestShapePoint_Backend(m, 0, 1);
}
int GetStrongestPositiveShapePoint(Mass * m)
{
  return GetStrongestShapePoint_Backend(m, +1, 1);
}
int GetStrongestNegativeShapePoint(Mass * m)
{
  return GetStrongestShapePoint_Backend(m, -1, 1);
}
int GetWeakestShapePoint(Mass * m)
{
  return GetStrongestShapePoint_Backend(m, 0, 0);
}
/* the returned shapepoint is always valid */
int GetFreeShapePoint(Mass * m)
{
  int i;
  for (i=0; i<MAX_SHAPEPOINTS; i++) {
    if (m->shapePoints[i].radius == 0 || m->shapePoints[i].strength == 0) {
      return i;
    }
  }
  return GetWeakestShapePoint(m);
}
double GetMassRadius(Mass * m)
{
  /* assumes a circle mass */
  return sqrt((double)(m->numPixels) / M_PI);
}
Point GetMassCenter(Mass * m)
{
  Point p;
  p.x = m->maxRect.x + m->maxRect.w/2;
  p.y = m->maxRect.y + m->maxRect.h/2;
  return p;
}
Point GetRandomPixelOfMass(Mass * m)
{
  Point p;
  assert(m->numPixels > 0);
  do {
    p.x = randarea(m->maxRect.x, m->maxRect.w);
    p.y = randarea(m->maxRect.y, m->maxRect.h);
  } while (FieldPoint(p)->owner != m);
  return p;
}
void MovePointRandomly(Point * p)
{
  if (randarea(0, 2)) {
    if (randarea(0, 2)) p->x++; else p->x--;
  } else {
    if (randarea(0, 2)) p->y++; else p->y--;
  }
}
void MovePointToPoint(Point * p, Point target)
{
  int deltax, deltay;
  deltax = abs(target.x - p->x);
  deltay = abs(target.y - p->y);
  if (deltax == 0 && deltay == 0) return;
  if (deltay == 0 || randarea(0, deltax+deltay) < deltax) {
    if (p->x < target.x) p->x++;
    if (p->x > target.x) p->x--;
  } else {
    if (p->y < target.y) p->y++;
    if (p->y > target.y) p->y--;
  }
}
void MovePointAwayFromPoint(Point * p, Point target)
{
  int deltax, deltay;
  deltax = abs(target.x - p->x);
  deltay = abs(target.y - p->y);
  if (deltax == 0 && deltay == 0) {
    MovePointRandomly(p);
    return;
  }
  if (deltay == 0 || randarea(0, deltax+deltay) < deltax) {
    if (p->x < target.x) p->x--;
    if (p->x > target.x) p->x++;
  } else {
    if (p->y < target.y) p->y--;
    if (p->y > target.y) p->y++;
  }
}
void MoveShapePointAwayFromOthers(Mass * m, int shapePoint)
{
  ShapePoint * target = &m->shapePoints[shapePoint];
  int best = -1;
  int best_value = -1;
  ShapePoint sp;
  int i;
  for (i=0; i<MAX_SHAPEPOINTS; i++) {
    sp = m->shapePoints[i];
    if (i != shapePoint && sp.radius >= 0 && sp.strength != 0) {
      int value;
      /* FIXME: is Wrap_ really necessary? */
      value = Wrap_GetCompareDistance(sp.pos, target->pos);
      if (best == -1 || value < best_value) {
	best_value = value;
	best = i;
      }
    }
  }
  if (best == -1) {
    MovePointRandomly(&target->pos);
  } else {
    MovePointAwayFromPoint(&target->pos, m->shapePoints[best].pos);
  }
}
Point GetBorderPointFromPointInDirection(Mass * m, Point p, int dir)
{
  /* this is in fact a poor implementation of a line drawing algo */
  int i;
  Point res, add;
  add = AngleToVector(dir, 256/4);
  res.x = p.x * 256; res.y = p.y * 256;
  for (i=0; i<1000; i++) {
    res.x += add.x;
    res.y += add.y;
    if (FieldXY(res.x/256, res.y/256)->owner != m) {
      res.x /= 256;
      res.y /= 256;
      return res;
    }
  }
  /* this would be a very big mass, should never happen. */
  assert(0);
  return res;
}
Point GetBorderPointFromCenterInDirection(Mass * m, int dir)
{
  return GetBorderPointFromPointInDirection(m, GetMassCenter(m), dir);
}
Point GetMassPixel(Mass * m)
{
  /* get a random pixel belonging to the mass */
  int pos, target;
  int x, y;
  pos = -1;
  target = randarea(0, m->numPixels);
  for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
    for (x=m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
      if (FieldXY(x, y)->owner == m) {
	if (++pos == target) {
	  Point p;
	  p.x = x;
	  p.y = y;
	  return p;
	}
      }
    }
  }
  {
    /* FIXME: assertion? */
    Point p;
    p.x = -1;
    p.y = -1;
    return p;
  }
}
Point AddBorderPixel(Mass * m)
{
  int trials = 500;
  Point p;
  p = GetMassCenter(m);
  if (FieldPoint(p)->owner != m) {
    /* live is not always that easy... find a pixel. */
    p = GetMassPixel(m);
  }
  while (FieldPoint(p)->owner == m && --trials) {
    Point np;
    np = p;
    switch (randarea(0, 4)) {
    case 0: np.x++; break;
    case 1: np.x--; break;
    case 2: np.y++; break;
    case 3: np.y--; break;
    }
    if (FieldPoint(np)->owner == m || !FieldPoint(np)->owner) {
      p = np;
    }
  }
  if (!trials) {
    p.x = -1; p.y = -1;
  } else {
    FieldPoint(p)->owner = m;
    FieldPoint(p)->height = HEIGHT_NORM/2;
    ExpandRectToIncludePoint(&m->maxRect, p.x, p.y);
    m->numPixels++;
  }
  return p;
}
void DeleteRandomPixel(Mass * m)
{
  int trials = 500;
  FieldPoint * p;
  DebugCheckMass(m);
  while (--trials) {
    p = FieldXY(randarea(m->maxRect.x, m->maxRect.w), randarea(m->maxRect.y, m->maxRect.h));
    if (p->owner == m) break;
  }
  if (trials) {
    p->owner = NULL;
    m->numPixels--;
  } else {
    DebugLevel(1, "No border point found to delete (random selection)\n");
  }
}
int CountPlants(Point p, int radius)
{
  /* OPTIMIZE perhaps using GetMassesInRect */
  /* OPTIMIZE this may be obsolete, use the interfaces instead */
  int x, y;
  int numMatches = 0;
  Mass * matchList[MAX_MASSES];
  Mass * m;
  for (y=p.y-radius; y<=p.y+radius; y++) {
    for (x=p.x-radius; x<=p.x+radius; x++) {
      if (FieldXY(x, y)->owner && FieldXY(x, y)->owner->isPlant) {
	int j;
	m = FieldXY(x, y)->owner;
	for (j=0; j<numMatches; j++) if (matchList[j] == m) break;
	if (j == numMatches) {
	  matchList[numMatches++] = m;
	}
      }
    }
  }
  return numMatches;
}
int CountFruits(Point p, int radius)
{
  int x, y;
  int numMatches = 0;
  Mass * matchList[MAX_MASSES];
  Mass * m;
  for (y=p.y-radius; y<=p.y+radius; y++) {
    for (x=p.x-radius; x<=p.x+radius; x++) {
      if (FieldXY(x, y)->owner && FieldXY(x, y)->owner->playerFood) {
	int j;
	m = FieldXY(x, y)->owner;
	for (j=0; j<numMatches; j++) if (matchList[j] == m) break;
	if (j == numMatches) {
	  matchList[numMatches++] = m;
	}
      }
    }
  }
  return numMatches;
}
//int LookAheadGetObstacle(Point p, int dir, int maxdist, Mass * ignoreMass)
/* return how empty the field look in this direction */
int LookAheadGetEmptyness(Point p, int dir, int maxdist, Mass * ignoreMass)
{
  /* OPTIMIZE: this is in fact a poor implementation of a line drawing
     algo */
  /* FIXME: in fact it's an evil code duplication, see above */
  int i;
  Point res, add;
  FieldPoint * fp;
  add = AngleToVector(dir, 256);
  res.x = p.x * 256; res.y = p.y * 256;
  for (i=maxdist; i--;) {
    res.x += add.x;
    res.y += add.y;
    fp = FieldXY(res.x/256, res.y/256);
    if (fp && fp->owner != ignoreMass) {
      return -(i*i);
    }
  }
  return 0;			/* the maximum */
}
/* return the better direction of dir+turn and dir-turn */
int LazyGetBetterDirection(Point start, int dir, int turn, int maxdist, Mass * ignoreMass)
{
  int a, b;
  a = LookAheadGetEmptyness(start, dir+turn, maxdist, ignoreMass);
  b = LookAheadGetEmptyness(start, dir-turn, maxdist, ignoreMass);
  if (a==b) return dir;
  return (a>b)?(dir+turn):(dir-turn);
}
/* look around (radial), and return the most empty direction */
Point LookAroundSimplestPath(Point start, int maxdist, Mass * ignoreMass)
{
  /* a vector that will point in the desired direction */
  /* Algo will generally not work - it will fail in this case:
      ooo   o
     o x-> ooo
      ooo   o
     And point stright into the obstacle.
  */
  Point res;
  int i;
  res.x = 0; res.y = 0;
  for (i=0; i<16; i++) {
    int emptyness;
    Point add;

    /* TODO: look ahead the line function */
    emptyness = LookAheadGetEmptyness(start, i*16, maxdist, ignoreMass);

    add = AngleToVector(i*16, emptyness);
    res.x += add.x/128;
    res.y += add.y/128;
  }
  //printf("res={%d, %d}\n", res.x, res.y);
  return res;
}
/* returns the number of pixels cut out */
int CutCircleFromMass(Mass * src, Mass * dst, Point center, int radius, int minCut)
{
  Rect r;
  int x, y;
  int dist;
  int cutPixels = 0;
  Point * points;
  r.x = center.x - radius;
  r.y = center.y - radius;
  r.w = 2*radius;
  r.h = 2*radius;
  points = malloc(r.w * r.h * sizeof(Point));
  radius = sqr(radius);
  for (y=r.y; y<r.y+r.h; y++) {
    for (x=r.x; x<r.x+r.w; x++) {
      dist = GetCompareDistance(center, XYToPoint(x, y));
      if (dist < radius && FieldXY(x, y)->owner == src) {
	points[cutPixels++] = XYToPoint(x, y);
      }
    }
  }
  if (cutPixels > minCut) {
    int i = cutPixels;
    while (i--) {
      FieldPoint(points[i])->owner = dst;
      Wrap_ExpandRectToIncludePoint(&dst->maxRect, points[i].x, points[i].y);
      src->numPixels--;
      dst->numPixels++;
    }
    return cutPixels;
  } else {
    return 0;
  }
}

void PreFlattening_BaseLineGrowth(Mass * m)
{
  /* This draws a line going through the center of the mass, in the
     m->heading direction (forth as well as back). Along this line,
     the height will increased/set. */
  Point center;
  Point pos;
  Point delta;
  FieldPoint * p;
  FieldPoint * lastPoint;
  Rect rect256;
  int backward;
  center = GetMassCenter(m);
  p = FieldPoint(center);
  /* not yet found */
  m->baseLineGrowth->head.x = -1;
  m->baseLineGrowth->tail.x = -1;
  /* everything is *256, to be more accurate but still fast. */
  /* + 128 for rounding */
  center.x = center.x * 256 + 128; 
  center.y = center.y * 256 + 128;
  rect256.x = m->maxRect.x * 256;
  rect256.y = m->maxRect.y * 256;
  rect256.w = m->maxRect.w * 256;
  rect256.h = m->maxRect.h * 256;
  delta = AngleToVector(m->heading, m->baseLineGrowth->step);
  if (m->baseLineGrowth->randomStepStart) {
    int movement;
    movement = randarea(-m->baseLineGrowth->step/2, m->baseLineGrowth->step);
    center.x += movement;
    center.y += movement;
  }
  pos = center;
  lastPoint = p;

  for (backward=0; backward<2; backward++) {
    if (backward) {
      /* restart at the center, but center itself is already done */
      delta.x = -delta.x;
      delta.y = -delta.y;
      pos.x = center.x + delta.x;
      pos.y = center.y + delta.y;
      p = FieldXY(pos.x>>8, pos.y>>8);
      lastPoint = p;
    }
    while (IsPointInRect(pos, rect256)) {
      /* 1. operate on this point (if valid) */
      if (p->owner == m) {
	if (m->baseLineGrowth->height) {
	  if (m->baseLineGrowth->setHeight) {
	    m->bufferMass += p->height - m->baseLineGrowth->height;
	    p->height = m->baseLineGrowth->height;
	  } else {
	    p->height += m->baseLineGrowth->height;
	    m->bufferMass -= m->baseLineGrowth->height;
	  }
	}
	lastPoint = p;
      }
      /* 2. go to next point */
      pos.x += delta.x;
      pos.y += delta.y;
      p = FieldXY(pos.x>>8, pos.y>>8);
    }
    /* special threatement of the start/end */
    {
      int height;
      if (backward) {
	height = m->baseLineGrowth->start;
	m->baseLineGrowth->head = FieldToPoint(lastPoint);
      } else {
	height = m->baseLineGrowth->end;
	m->baseLineGrowth->tail = FieldToPoint(lastPoint);
      }
      if (height > 0) {
	p = lastPoint;
	if (m->baseLineGrowth->setHeight) {
	  m->bufferMass += p->height - height;
	  p->height = height;
	} else {
	  p->height += height;
	  m->bufferMass -= height;
	}
      }
    }
  }
}

void PreFlattening_BaseLineGrowthFuzzy(Mass * m)
{
  int x, y; 
  int dx, dy;
  int distance;
  Point center;
  center = GetMassCenter(m);

  dy = cos((float)m->heading * 2 * M_PI / 256) * 256;
  dx = sin((float)m->heading * 2 * M_PI / 256) * 256;
  /* that's what it is at the center point */
  distance = 0;			
  /* go to the starting point */
  distance += (m->maxRect.x - center.x) * dx;
  distance += (m->maxRect.y - center.y) * dy;
  for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
    for (x=m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
      int h;
      FieldPoint * p;
      p = FieldXY(x, y);
      if (p->owner == m) {
	h = (distance*distance)/(256*256)*HEIGHT_NORM/8;
	assert(h>=0);
	if (h > p->height) {
	  m->bufferMass += p->height;
	  p->height = 0;
	} else {
	  m->bufferMass += h;
	  p->height -= h;
	}
      }
      distance += dx;
    }
    distance -= m->maxRect.w * dx;
    distance += dy;
  }
}

#if 0
/* TODO: find a good location for such things. */
void MazeWallChecks(Mass * m)
{
  int i, over;
  Point move;
  Point c1;
  move.x = 0;
  move.y = 0;
  c1 = GetMassCenter(m);
  over = 0;
  { /* 1. move into the direction of interfaced but unconnected walls. */
    for (i=0; i<numInterfaces; i++) {
      assert(interfaces[i].m[0] != interfaces[i].m[1]);
      if (interfaces[i].m[1] == m) {
	interfaces[i].m[1] = interfaces[i].m[0];
	interfaces[i].m[0] = m;
      }
      if (interfaces[i].m[0] == m && /*interfaces[i].m[0]->isMazeWall &&*/ !interfaces[i].contact ) {
	Mass * m2 = interfaces[i].m[1];
	/* the int */
	move = Wrap_VectorFromTo(c1, GetMassCenter(m2));
	over = 1;
      }
    }
  }

  if (!over) {
    /* 2. keep a minimum distance from unconnected masses nearby */
    int radius = 40;
    int x, y;
    Mass * list[50];
    Mass ** p;
    list[0] = NULL;
    for (y=c1.y-radius; y<c1.y+radius; y++) {
      Mass * m2 = FieldXY(x, y)->owner;
      if (m2) {
	/* add the item to the list, and avoid double entries */
	int found = 0;
	for (p=list; *p; p++) {
	  if (*p == m2) {
	    found = 1;
	    break;
	  }
	} 
	if (!found) {		/* FIXME: possible overflow (50) */
	  *(p++) = m2;
	  *(p) = NULL;
	}
      }
    }
    /* now the list contains all masses in the neighbourhood */
    for (p=list; *p; p++) {
      Point v;
      int d;
      Mass * m2;
      m2 = *p;
      d = Wrap_GetRealDistance(c1, GetMassCenter(m2));
      if (d < 4*GetMassRadius(m) /*OPTIMIZE!!! sqrt... FIXME: other
				   way to measure distance (border...) */) {
	/* too near - move away */
	v = Wrap_VectorFromTo(c1, GetMassCenter(m2));
      } else {
	/* too far - move towards it (only if it is another wall??) */
	v = Wrap_VectorFromTo(GetMassCenter(m2), c1);
	/* TODO: only if not too far away. */
      }
      move.x += (double)(v.x) / GetRealLength(v) * 256;
      move.y += (double)(v.y) / GetRealLength(v) * 256;
      over = 1;
    }
  }
  if (!over) {
    /* move randomly.. or better, not at all. */
    m->flatteningMovement = 0;
    return;
  }
  
  /* calculate final moving direction */
  m->heading = VectorToAngle(move);
  m->flatteningMovement = 1;
}
#endif

void NexusPreFlattening(Mass * m)
{
  int i;
  /* control size
  if (m->numPixels < 40 * m->nexusLevel) {
    m->bufferMass += HEIGHT_NORM * m->numPixels / 16;
  } */
  if (m->nexusLevel == 5 /*&& m->age < 400*/) {
    m->bufferMass += 10*HEIGHT_NORM;
  }
  if (m->nexusLevel == 4 && m->age % 8 == 0 && m->energy > 0) {
    m->heading++;
  }
  if (m->nexusLevel == 4 && m->age > 3000) m->bufferMass -= HEIGHT_NORM;
  if (m->nexusLevel == 3) m->bufferMass -= 8*HEIGHT_NORM;

  if (m->age < 200 && m->nexusLevel < 5) {
    if (m->age % 16 == 0) {
      for (i=0; i<3; i++) m->baseColor[i] = MAX(0, m->baseColor[i]-1);
    }
    return;
  }

  m->energy++;

  if (m->nexusLevel == 5 && m->numPixels > 200 && m->age % 64 == 0 && m->energy > 0) {
    /* birth of a new mass */
    Mass * m2;
    Point pos;
    int numCut;
    //int dir;
    m->energy = 0;
    //for (dir = 0; dir < 255; dir += 32) {
    m->heading += 40;
    m2 = CreateMass();
    memcpy(m2->baseColor, m->baseColor, sizeof(Color));
    m2->nexusLevel = m->nexusLevel - 1;
    m2->velocity = 256;
    m2->flatteningMovement = 1;
    m2->heading = m->heading;
    pos = GetBorderPointFromCenterInDirection(m, m2->heading);
    numCut = CutCircleFromMass(m, m2, pos, 5, 10);
    if (!numCut) DestroyMass(m2);
    DebugLevel(2, "Cut %d pixels from NexusMass\n", numCut);
    //}
  }

  /* The nexus sends everybody out! */
  if (m->energy > 8000 /* todo... looks ugly. */) {
    Mass ** mm;
    Mass * m2;
    Rect r;
    Point center;
    int radius = 200;
    m->energy = -2000;
    center = GetMassCenter(m);
    r = m->maxRect;
    r.x -= radius;
    r.y -= radius;
    r.w += 2*radius;
    r.h += 2*radius;
    mm = GetMassesInRect(r);
    for (i=0; mm[i]; i++) {
      m2 = mm[i];
      if (m2->nexusLevel < 5) {
	Point p;
	p = GetMassCenter(m2);
	if (Wrap_GetCompareDistance(center, p) < sqr(radius)) {
	  m2->heading = VectorToAngle(Wrap_VectorFromTo(center, p));
	  m2->energy = -800;
	}
      }
    }
    free(mm);
  }

  if (m->nexusLevel == 4) {
    /* a descending branch */
    //m->playerFood = 1;
    //if (m->baseColor[0] < 255) m->baseColor[0]++;
  }
}

int PreFlattening(Mass * m)
{
  /* FIXME: split this up into per-mass-type or per-feature functions. */

  m->age++;

  if (m->nexusLevel) NexusPreFlattening(m);
  
  if (m->player != -1) {
    /* move the player! */
#if 0
    /* shapepoint movement */
    ShapePoint * sp;
    Point newpos;
    sp = &m->shapePoints[0];
    newpos = sp->pos;
    newpos.x += player[m->player].movex;
    newpos.y += player[m->player].movey;

    if (FieldPoint(newpos)->owner != m) {
      newpos = sp->pos;
      if (FieldXY(newpos.x, newpos.y)->owner != m) {
	newpos.x = m->maxRect.x + m->maxRect.w/2;
	newpos.y = m->maxRect.y + m->maxRect.h/2;
      }
    }
    sp->pos = newpos;
    if (sp->strength < HEIGHT_NORM/5) sp->strength += HEIGHT_NORM/50;
    if (sp->strength > HEIGHT_NORM/5) sp->strength -= HEIGHT_NORM/50;
    if (sp->radius < 12) sp->radius++;
    if (sp->radius > 12) sp->radius--;
#else
#ifdef PLAYER_MOVE_NIBBLES
    m->flatteningMovement = 1;
    m->heading += 1*player[m->player].movex;
    /*
    m->velocity = -256*player[m->player].movey;
    */
    m->velocity += -player[m->player].movey * 3;
    m->velocity -= sign(m->velocity);
#else
    Point v;
    v.x = player[m->player].movex;
    v.y = player[m->player].movey;
    if (v.x == 0 && v.y == 0) {
      m->velocity = 0;
      //m->flatteningMovement = 0;
    } else {
      m->flatteningMovement = 1;
      m->heading = VectorToAngle(v);
      m->velocity = 256;
    }
#endif
#endif
    /* TODO: shrink the player? */
    /* TODO: adapt speed to size? */
    /* drop bombs */
    if (m->energy < 256) m->energy++;
    if (player[m->player].button1) {
      Point c;
      c = GetMassCenter(m);
      while (m->energy > 10) {
	m->energy -= 10;
	CreateMassattackMass(c);
      }
    }
#if 0
    if (player[m->player].button1 && m->energy >= 200) {
      float dir;
      Point p;
      /* TODO: choose a predictable location */
      dir = (float)randarea(-32, 65) / 64.0 * M_PI;
      p = GetBorderPointFromCenterInDirection(m, dir);
	Mass * bomb;
	if (bomb = CreateBombMass(p.x, p.y)) {
	  m->energy = 0;
	}
    }
#endif /* 0 */
    if (m->player == 0) {
      displayHeading = m->heading;
      DebugLevel(3, "DisplayHeading = %d\n", displayHeading);
    }
  }

  if (m->cell && m->age > 10) {
    if (m->energy < 40) m->energy++;
    if (m->cell->pulse > 1) m->cell->pulse--;
    if (m->velocity < 156) m->velocity += 10;
    if (m->age % 4 == 0) {
      /* time to reconsider the situation */
      int i, desiredAngle, totPulse;
      Mass ** ml;
      Mass * m2;
      Mass * nearest;
      int nearestDist;
      int sightDistance = 60;
      Rect r;
      Point center;
      Point desiredVector;
      /* FIXME: doesn't work as expected */
      //desiredAngle = VectorToAngle(LookAroundSimplestPath(GetMassCenter(m), 40, m));
      center = GetMassCenter(m);

      /* find all other swarm masses in region */
      r.x = center.x - sightDistance/2;
      r.y = center.x - sightDistance/2;
      r.w = sightDistance;
      r.h = sightDistance;
      ml = GetMassesInRect(r);
      totPulse = 0;
      desiredAngle = 0; // keep the compiler happy
      nearestDist = 0;  // dito
      desiredVector.x = 0;
      desiredVector.y = 0;
      nearest = NULL;
      for (i=0; (m2=ml[i]); i++) {
	if (m2->cell) {
	  int d;
	  desiredVector = AddVectors(desiredVector, AngleToVector(m2->heading, m2->cell->pulse * 256));
	  totPulse += m2->cell->pulse * 256;
	  d = Wrap_GetCompareDistance(center, GetMassCenter(m2));
	  if (d < nearestDist || !nearest) {
	    nearest = m2;
	    nearestDist = d;
	  }
	}
      }
      free(ml);
      /* TODO: idea: the swarm seekes players. It will circle around
	 one, because if it comes to close the pressure to avoid the
	 obstacle (the player mass) becomes more important than going
	 stright to the player, and it will stay circling because it
	 can not adapt the new direction (the player) fast enough (or
	 just as before, doesn't want to get to close to the obstacle)
      */
      if (totPulse) {
	/* the desire is all about peer pressure ;-) */
	if (nearestDist > sqr(sightDistance*2/4)) {
	  /* oh, oh, the others are going to get out of sight! */
	  //if (m->speed < 2*256) m->speed += 25;
	  /* cry out in pain! */
	  m->cell->pulse += 20;
	  m->energy -= 20;
	  if (m->energy < 0) {
	    m->cell->pulse += m->energy;
	    m->energy = 0;
	  }
	  //desiredAngle = VectorToAngle(VectorFromTo(center, GetMassCenter(nearest)));
	  //m->heading = VectorToAngle(VectorFromTo(center, GetMassCenter(nearest)));
	  //m->heading += 20*sign(((desiredAngle - m->heading)&255) - 128);
	} else {
	  if (m->speed > 180) m->speed -= 25;
	  desiredAngle = VectorToAngle(desiredVector);
	  m->heading -= 10*sign(((desiredAngle - m->heading)&255) - 128);
	}
      }
      /* but without the will of each member, a swarm does nothing */
      m->heading = LazyGetBetterDirection(center, m->heading, 5, 40, m);
      //m->heading += randarea(-10, 21);
    }
    /* show the pulse */
    RGBSET(m->baseColor, MIN(255, 100+m->cell->pulse), MIN(255, 100+m->cell->pulse), MIN(255, m->cell->pulse*10));
  }

#ifdef OLD_GAME_OF_LIVE_CODE
  if (m->cell) {
    if (m->age > 8) {
      int r;		/* radius */
      int found;
      Mass * neigh[8];
      Point center;
      center = GetMassCenter(m);
      found = 0;
      r = 0;
      m->age = 0;
      /* get the eight nearest cell masses */
      /* FIXME: it is not circular but a rectangle */
      while (found < 8 && r < 10) {
	int steps;
	int i;
	r++;
	for (i=0; i<4; i++) {
	  for (steps=-r; steps <= r; steps++) {
	    int x, y;
	    Mass * m2;
	    switch (i) {
	    case 0:
	      x = center.x + r;
	      y = center.y + steps;
	      break;
	    case 1:
	      x = center.x - r;
	      y = center.y + steps;
	      break;
	    case 2:
	      x = center.x + steps;
	      y = center.y + r;
	      break;
	    case 3:
	      x = center.x + steps;
	      y = center.y - r;
	      break;
	    default:
	      assert(0);
	    }
	    if ((m2 = FieldXY(x, y)->owner) && (m2->cell) && (m != m2)) {
	      int j, ok;
	      ok = 1;
	      for (j=0; j<found; j++) if (neigh[j] == m2) ok = 0;
	      if (ok) {
		neigh[found++] = m2;
	      }
	    }
	    if (found == 8) break;
	    assert(found < 8);
	  }
	  if (found == 8) break;
	  assert(found < 8);
	}
      }
      if (found == 8) {
	int i;
	int c = 0;
	for (i=0; i<8; i++) {
	  c += neigh[i]->cell->golState;
	}
	/* game of live rules */
	if ((m->cell->golState == 1 && (c == 2 || c == 3))
	    || (m->cell->golState == 0 && (c == 3))) {
	  m->cell->golState = 1;
	} else {
	  m->cell->golState = 0;
	}
	if (m->cell->golState == 1) {
	  RGBSET(m->baseColor, 220, 220, 255);
	} else {
	  RGBSET(m->baseColor, 80, 80, 80);
	}
      } else {
	/* did not find enough neighbours to play the game - do not
	   change the state. */
      }
    }
  }
#endif /* OLD_GAME_OF_LIVE_CODE */

  if (m->isBomb) {
    if (m->age > 200) {
      int i;
      int dir;
      Point p;
      /* explode */
      dir = randarea(-64, 129);
      for (i=0; i<4; i++) {
	Mass * pj;
	dir += 64;
	p = GetBorderPointFromCenterInDirection(m, dir);
	pj = CreateProjectilMass(p.x, p.y);
	if (pj) {
	  pj->heading = dir;
	}
      }
      DestroyMass(m);
      return 1;
    }
  }

  if (m->isProjectil) {
    if (m->age > 20) {
      if (m->vivid < 200) m->vivid++;
    }
  }

  if (m->isMeltingMass) {
    if (m->age > 200) {
      m->isMeltingMass = 0;
      m->isPlant = 1;
      m->flatteningMovement = 0;
      m->vivid = 2;
    }
  }

  if (m->isWall && m->age <= 30) {
    if (m->age == 30) m->speed = 25;
    m->bufferMass += m->numPixels * HEIGHT_NORM;
  }

  if (m->isPlant) {
    if (m->age < 190) {
      /* let the plant grow */
      m->bufferMass += HEIGHT_NORM * 2;
    } else if (m->age > 2048) {
      /* TODO: let the plant die? */
    } else {
      /* the plant is fully developed */
      if (randarea(0, 20) == 0) {
	/* do less frequent checks */
	/* TODO (whishlist):
	   - check other plants nearby, environement
	   - check mass contacts into heading direction
	   - create a new branch
	   - create a new fruit
	*/
	if (CountPlants(GetMassCenter(m), 30) < 9) {
	  if (randarea(0, 10) == 0) {
	    /* add a new branch (perhaps.) */
	    float direction;
	    Point p;
	    Mass * m1;
	    direction = randarea(-16, 33);
	    direction += m->heading;
	    /* TODO: not from center... or, not always */
	    p = GetBorderPointFromCenterInDirection(m, direction);
	    /* FIXME: why on earth would this need a cast, as the comipler warns? */
	    m1 = CreateGrowingFoodMass(p.x, p.y);
	    if (m1) {
	      m1->heading = direction;
	      /*
		for (i=1; i<4; i++) {
		m1->shapePoints[i].pos.x = p.x + (int)(cos(direction)*2*i);
		m1->shapePoints[i].pos.y = p.y + (int)(sin(direction)*2*i);
		m1->shapePoints[i].radius = 2;
		m1->shapePoints[i].strength = HEIGHT_NORM/2;
		p = m1->shapePoints[i].pos;
		if ((i == 3) && FieldPoint(p)->owner && FieldPoint(p)->owner->isPlant) {
		DestroyMass(m1);
		break;
		}
		}
	      */
	    }
	  }
	} else {
	  /* plant will not grow further. */
	  /* perhaps grow a fruit. */
	  if (randarea(0, 40) == 0 && CountFruits(GetMassCenter(m), 20) < 2) {
	    /* hmm... turn into a fruit. why not. */
	    /* TODO: make it smoother */
	    m->isPlant = 0;
	    m->playerFood = 1;
	    m->age = 0;
	    if (m->speed > 10) m->speed /= 10;
	    //RGBSET(m->baseColor, 255, 111, 80);
	  }
	}
      }
    }
  } /* if (m->isPlant) */

  if (m->isMeltingMass) {
    /* move it into heading direction */
  }

  if (m->isMazeWall) {
    //MazeWallChecks(m);
  }
    
  if (m->playerFood) {
    if (m->baseColor[0] < 255) m->baseColor[0]++;
    if (m->baseColor[1] > 111) m->baseColor[1]--;
    if (m->baseColor[2] > 80) m->baseColor[2]--;
  }

  if (m->vivid > 0) {
    int r, i;
    Point p;
    r = sqrt((float)(m->numPixels) / M_PI) * 2;
    if (r < 2) r = 2;
    m->heading += randarea(-2, 5);
    p = GetBorderPointFromCenterInDirection(m, m->heading+128);
    for (i=0; i<m->vivid; i++) {
      Point p1;
      FieldPoint * fp;
      p1.x = p.x + randarea(-r/2, r);
      p1.y = p.y + randarea(-r/2, r);
      if ((fp=FieldPoint(p1))->owner == m) {
	m->bufferMass += fp->height/2;
	fp->height -= fp->height/2;
      }
    }
  }

  if (m->vivid2) {
    /* create a "bearable" number of avoidpoints */
    int i, count, idealCount;
    ShapePoint * sp;
    idealCount = m->numPixels / 9;
    count = CountShapePoints(m);
    if (count < idealCount) {
      sp = &m->shapePoints[GetFreeShapePoint(m)];
      sp->radius = 7;
      sp->strength = m->vivid2*HEIGHT_NORM/4;
      sp->pos.x = randarea(m->maxRect.x, m->maxRect.w);
      sp->pos.y = randarea(m->maxRect.y, m->maxRect.h);
    }
    /* move them randomly */
    for (i=0; i<MAX_SHAPEPOINTS; i++) {
      sp = &m->shapePoints[i];
      sp->pos.x += randarea(-1, 3);
      sp->pos.y += randarea(-1, 3);
      /* perhaps... FitPointIntoCircle? */
      Wrap_FitPointIntoRect(&sp->pos, m->maxRect);
    }
  }
    
  PreFlattening_ShapePoints(m);

  if (m->baseLineGrowth) PreFlattening_BaseLineGrowth(m);

  if (m->isPlant2) {
    // m->heading++;
    /*
      - an der Spitze wachsen, aber in Kontakt mit dem Stamm bleiben
      - Richtung (heading) veraendern (freies Gelaende bevorzugt?)
      - wenn ein Hinderniss kommt *nicht* wachsen.
      - Wenn Wachstum abgeschlossen, eventuell einen Ast wachsen lassen
      - Verlangsamung (speed), erhaertung.
      - Ueberpruefen, ob Kontakt zum Stamm. Wenn nein, absterben.
    */
    if (!m->growRounds) {
      //m->baseLineGrowth->height = HEIGHT_NORM/2;
      m->bufferMass += HEIGHT_NORM;
      //m->baseLineGrowth->height = HEIGHT_NORM/8;
      PreFlattening_BaseLineGrowthFuzzy(m);
      //m->baseLineGrowth->end = HEIGHT_NORM*4;
    }
    m->heading++;
  }

  if (m->isCoinMass) {
    DeleteRandomPixel(m);
    AddBorderPixel(m);
  }
  
  if (m->player != -1) {
    /* re-center display */
    fieldCenter = GetMassCenter(m);
  }

  return 0;
}

void DebugMaxRectCheck(Mass * m)
{
#ifdef DEBUG_HEAVY
  int x, y;
  for (y=0; y<FIELDSIZE; y++) for (x=0; x<FIELDSIZE; x++) {
    /* we have to take care of rectangles that are wrapped around */
    /* TODO: move this code into a wraprectcheck funtion */
    int x1, y1;
    x1 = (x - m->maxRect.x) & FIELDWRAP;
    y1 = (y - m->maxRect.y) & FIELDWRAP;
    if (x1 >= 0 && x1 < m->maxRect.w && y1 >= 0 && y1 < m->maxRect.h) { 
      // nothing
    } else {
      /* this point is not included in the maxRect */
      assert(FieldXY(x, y)->owner != m);
    }	
  }
#else
  /* TODO: replace this code, it's not that easy with wrapping borders
     any more */
  /*
  int x, y;
  for (x=m->maxRect.x-2; x<m->maxRect.x+m->maxRect.w+2; x++) {
    y = m->maxRect.y-1;
    assert(FieldXY(x, y)->owner != m);
    y = m->maxRect.y-2;
    assert(FieldXY(x, y)->owner != m);
  }
  for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
    x = m->maxRect.x-1;
    assert(FieldXY(x, y)->owner != m);
    x = m->maxRect.x+m->maxRect.w;
    assert(FieldXY(x, y)->owner != m);
  }
  */
#endif
}

MassInterface * FindMassInterface(Mass * m, Mass * m2)
{
  MassInterface * mi = m->interfaces;
  while (mi && mi->m != m2) mi = mi->next;
  assert(!mi || mi->m == m2);
  // debugging only... remove it if NULL is a desired return value
  assert(m == m2 || mi);
  return mi;
}

/* good code */
void ConvolveMass256_nonexpanding(Mass * m, int * matrix, int radius)
{
  int x, y;
  int * copy;
  int * src;
  Rect r;
  if (m->numPixels == 0) return;

  r = m->maxRect;
  r.x -= radius;
  r.y -= radius;
  r.w += 2*radius;
  r.h += 2*radius;

  copy = malloc(r.h*r.w*sizeof(int));

  /* selectively copy the source region */
  src = copy;
  for (y=r.y; y<r.y+r.h; y++) {
    for (x=r.x; x<r.x+r.w; x++) {
      FieldPoint * p;
      int h;
      p = FieldXY(x, y);
      if (!p->owner) {
	h = 0;
      } else if (p->owner == m) {
	h = p->height;
      } else {
	MassInterface * mi;
	mi = FindMassInterface(m, p->owner);
	h = (p->height * mi->borderReaction) >> 8;
      }
      *src++ = h;
    }
  }
  assert(src == copy+r.h*r.w);

  /* apply the matrix using the copy as input */
  {
    int size;
    int wraparound;
    int buffdiff=m->bufferMass;

    size = 2*radius+1;
    src = copy;
    wraparound = r.w - size;
    for (y=r.y; y<r.y+r.h; y++) {
      for (x=r.x; x<r.x+r.w; x++) {
	FieldPoint * p;
	p = FieldXY(x, y);
	if (p->owner == m) {
	  int i, j;
	  int * mt;
	  int h = 0;
	  mt = matrix;
	  for (i=size; i--;) {
	    for (j=size; j--;) {
	      h += *src++ * *mt++;
	    }
	    src += wraparound;
	  }
	  if (h<0) h=0;
	  h >>= 8;
	  m->bufferMass += p->height - h;
	  p->height = h;
	  src = src - size * r.w;
	}
	src++;
      }
      src += size-1;
    }
    assert(src + 2*radius*r.w == copy + r.w*r.h);
    buffdiff = m->bufferMass - buffdiff;
    DebugLevel(1, "Buffdiff = %d (%d * HEIGHT_NORM)\n", buffdiff, buffdiff/HEIGHT_NORM);
  }
  
  free(copy);
}

/* faster movement */
void ConvolveMass256_expanding(Mass * m, int * matrix, int radius)
{
  int x, y;
  int * copy;
  int * src;
  int radius2;
  Rect r;
  if (m->numPixels == 0) return;

  radius2 = 2*radius;
  r = m->maxRect;
  r.x -= radius;
  r.y -= radius;
  r.w += 2*radius;
  r.h += 2*radius;

  copy = malloc(r.h*r.w*sizeof(int));

  /* selectively copy the source region */
  src = copy;
  for (y=m->maxRect.y-radius; y<m->maxRect.y+m->maxRect.h+radius; y++) {
    for (x=m->maxRect.x-radius; x<m->maxRect.x+m->maxRect.w+radius; x++) {
      FieldPoint * p;
      int h;
      p = FieldXY(x, y);
      if (!p->owner) {
	h = 0;
      } else if (p->owner == m) {
	h = p->height;
      } else {
	MassInterface * mi;
	mi = FindMassInterface(m, p->owner);
	h = (p->height * mi->borderReaction) / 256;
      }
      *src++ = h;
    }
  }
  assert(src == copy+r.h*r.w);

  /* apply the matrix using the copy as input */
  {
    int size;
    int wraparound;

    size = 2*radius+1;
    src = copy;
    wraparound = r.w - size;
    for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
      for (x=m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
	FieldPoint * p;
	p = FieldXY(x, y);
	/* old, good code:
	if (p->owner == m) {
	  int i, j;
	  int * mt;
	  int h = 0;
	  mt = matrix;
	  for (i=size; i--;) {
	    for (j=size; j--;) {
	      h += *src++ * *mt++;
	    }
	    src += wraparound;
	  }
	  if (h<0) h=0;
	  h >>= 8;
	  m->bufferMass += p->height - h;
	  p->height = h;
	  src = src - size * r.w;
	}
	*/
	/* new, fast moving, evil code: */
	{
	  int i, j;
	  int * mt;
	  int h = 0;
	  mt = matrix;
	  for (i=size; i--;) {
	    for (j=size; j--;) {
	      h += *src++ * *mt++;
	    }
	    src += wraparound;
	  }
	  if (h<0) h=0;
	  h >>= 8;
	  if (!p->owner && h > m->minBorderHeight) {
	    p->owner = m;
	    p->height = 0;
	    ExpandRectToIncludePoint(&m->maxRect, x, y);
	    m->numPixels++;
	  }
	  if (p->owner == m) {
	    m->bufferMass += p->height - h;
	    p->height = h;
	  }
	  src = src - size * r.w;
	}
	src++;
      }
      src += size-1;
    }
    assert(src + 2*radius*r.w == copy + r.w*r.h);
  }
  
  free(copy);
}

void UpdateInterfaces(Mass * m)
{
  /* FIXME: this code looks like a slaughter. */
  Mass ** neighbours;
  int i;
  Rect r;
  MassInterface * miOld;
  MassInterface * miNew;

  /* have to look at least one pixel ahead to find all masses that
     could be in touch with this one */
  assert(m->lookDistance >= 1);
  r.x = m->maxRect.x - m->lookDistance;
  r.y = m->maxRect.y - m->lookDistance;
  r.w = m->maxRect.w + 2*m->lookDistance;
  r.h = m->maxRect.h + 2*m->lookDistance;
  neighbours = GetMassesInRect(r);

  /* remove stale entries, update lasting entries */
  miOld = m->interfaces;
  miNew = NULL;
  while (miOld) {
    MassInterface * miNext = miOld->next;
    for (i=0; neighbours[i]; i++) if (neighbours[i] == miOld->m) break;
    if (neighbours[i] || miOld->m == NULL) {
      if (!miNew) {
	miNew = m->interfaces = miOld;
      } else {
	miNew->next = miOld;
	miNew = miNew->next;
      }
      miNew->oldContact = miNew->contact;
      miNew->contact = 0;
      miNew->next = NULL;
    } else {
      free(miOld);
    }
    miOld = miNext;
  }

  /* add new entries */
  for (i=0; neighbours[i]; i++) {
    MassInterface * mi;
    mi = m->interfaces;
    while (mi && mi->m != neighbours[i]) mi = mi->next;
    if (!mi) {
      mi = calloc(1, sizeof(MassInterface));
      if (!miNew) {
	miNew = m->interfaces = mi;
      } else {
	miNew->next = mi;
	miNew = miNew->next;
      }
      miNew->m = neighbours[i];
    }
  }

  free(neighbours);

  /* self-check */
  neighbours = GetMassesInRect(r);
  for (i=0; neighbours[i]; i++) assert(FindMassInterface(m, neighbours[i]));
  free(neighbours);
}

/* FIXME: rename this function, it actually does lots more of
   housekeeping */
void DistributeBufferMass(Mass * m)
{
  int x, y;
  int addToEachPixel;
  Rect r;
  if (m->numPixels == 0) return;

  /* OPTIMIZE: perhaps this could be done elsewhere */
  m->volume = 0;
  addToEachPixel = m->bufferMass / m->numPixels; /* may get < 0 */
  DebugLevel(4, "AddToEachPixel is %d (%d/%d)\n", addToEachPixel, m->bufferMass, m->numPixels);

  r = m->maxRect;
  m->maxRect.w = 0;
  m->maxRect.h = 0;

  for (y=r.y; y<r.y+r.h; y++) {
    for (x=r.x; x<r.x+r.w; x++) {
      FieldPoint * p;
      p = FieldXY(x, y);
      if (p->owner == m) {
	/* OPTIMIZE: do this once, not in the loop */
	m->bufferMass -= addToEachPixel;
	p->height += addToEachPixel;
	if (p->height < 0) {
	  m->bufferMass += p->height;
	  p->height = 0;
	}
	m->volume += p->height;
	ExpandRectToIncludePoint(&m->maxRect, x, y);
      }
    }
  }
}

void HandleBorderPoints(Mass * m)
{
  Rect r;
  int x, y;
  int i;
  /* TODO: try this with 8 directions, just for fun :) */
  const int addx[4] = { -1, 0, +1, 0 };
  const int addy[4] = { 0, -1, 0, +1 };
  DebugLevel(3, "HandleBorderPoints\n");
  /* FIXME: the m->maxRect changes during this algo. Would it be a
     good idae to do the for loop with those values directly? */
  r = m->maxRect;
  for (y=r.y; y<r.y+r.h; y++) {
    for (x=r.x; x<r.x+r.w; x++) {
      FieldPoint * p;
      p = FieldXY(x, y);
      if (p->owner == m) {
	if (p->height < m->minBorderHeight) {
	  p->owner = NULL;
	  m->bufferMass += p->height;
	  m->numPixels--;
	} else {
	  /* TODO: randomize which direction gets checked first */
	  FieldPoint * n[4];
	  int isBorderPoint = 4;

	  /* manually unrolled loop, optimized?:
	  for (i=4; i--;) {
	    n[i] = FieldXY(x+addx[i], y+addy[i]);
	    if (n[i]->owner == m) isBorderPoint--;
	  }
	  */
	  n[0] = FieldXY(x-1, y+0);
	  if (n[0]->owner == m) isBorderPoint--;
	  n[1] = FieldXY(x+0, y-1);
	  if (n[1]->owner == m) isBorderPoint--;
	  n[2] = FieldXY(x+1, y+0);
	  if (n[2]->owner == m) isBorderPoint--;
	  n[3] = FieldXY(x+0, y+1);
	  if (n[3]->owner == m) isBorderPoint--;

	  if (isBorderPoint) {
	    for (i=4; i--;) {
	      if (n[i]->owner != m && n[i]->owner) {
		MassInterface * mi;
		Mass * m2 = n[i]->owner;
		assert(IsValidMass(m2));
		mi = FindMassInterface(m, m2);
		mi->contact++;
		if (mi->eatme) {
		  p->owner = m2;
		  m->numPixels--;
		  m2->numPixels++;
		  Wrap_ExpandRectToIncludePoint(&m2->maxRect, x, y);
		  /*
		    mi->m->bufferMass += p->height;
		    p->owner = NULL;
		    m->numPixels--;
		  */
		  break;
		} else if (mi->borderReaction <= 0 && n[i]->height > p->height) {
		  p->owner = NULL;
		  m->bufferMass += p->height;
		  m->numPixels--;
		  break;
		}
	      }
	    }
	    if (p->owner == m && p->height > m->maxBorderHeight) {
	      for (i=4; i--;) {
		if (!n[i]->owner) {
		  n[i]->owner = m;
		  n[i]->height = p->height / 2;
		  m->bufferMass -= p->height / 2;
		  ExpandRectToIncludePoint(&m->maxRect, x+addx[i], y+addy[i]);
		  m->numPixels++;
		  break;
		}
	      }
	    }
	  }
	}
      }
    }
  }
}

void DoConvolveFlattening(Mass * m)
{
  /* light blur matrix
  int matrix2[3*3] = {
    1*16, 2*16, 1*16,
    2*16, 4*16, 2*16,
    1*16, 2*16, 1*16,
  }; */
  /* big gaussian blur matrix */
  int matrix[7*7] = {
    0,0 ,0  ,5  ,0  ,0 ,0,
    0,5 ,18 ,32 ,18 ,5 ,0,
    0,18,64 ,100,64 ,18,0,
    5,32,100,100,100,32,5,
    0,18,64 ,100,64 ,18,0,
    0,5 ,18 ,32 ,18 ,5 ,0,
    0,0 ,0  ,5  ,0  ,0 ,0,
  };
  /* just for testing, same as above
  int matrixBig[13*13] = {
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,5  ,0  ,0 ,0,0,0,0,
    0,0,0,0,5 ,18 ,32 ,18 ,5 ,0,0,0,0,
    0,0,0,0,18,64 ,100,64 ,18,0,0,0,0,
    0,0,0,5,32,100,100,100,32,5,0,0,0,
    0,0,0,0,18,64 ,100,64 ,18,0,0,0,0,
    0,0,0,0,5 ,18 ,32 ,18 ,5 ,0,0,0,0,
    0,0,0,0,0 ,0  ,5  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
    0,0,0,0,0 ,0  ,0  ,0  ,0 ,0,0,0,0,
  }; 
  */
#define RADIUS 3

  /* TODO: movement is a problem... it would be so easy... */
  /* ideas: gaussian blur constructed from scratch, with centre
     replaced according to the current speed. Use radius from center,
     and a gaussian curve, maybe indexed [256] */

  int x, y;
  /* FIXME: don't use double */
  int vx, vy;
  int sum = 0;
  Point p;
  p = AngleToVector(m->heading, m->velocity);
  vx = p.x; vy = p.y;
  for (y=-RADIUS; y<=RADIUS; y++) {
    for (x=-RADIUS; x<=RADIUS; x++) {
      int v;
      if (m->flatteningMovement && 0) {
	/* TODO: Rewrite. Think about an algo that works. */
	double val;
	val = vx*x + vy*y;
	val /= 4;
	v = val + 128;
	//DebugLevel(1, "v=%d\n", v);
	v = MIN(v, 255);
	v = MAX(v, 0);
	matrix[(y+RADIUS)*(2*RADIUS+1)+(x+RADIUS)] = v;
      } else {
	v = matrix[(y+RADIUS)*(2*RADIUS+1)+(x+RADIUS)];
      }
      sum += v;
    }
  }
  for (x=0; x<sqr(2*RADIUS+1); x++) matrix[x] = (matrix[x]*256+sum/2)/sum;
  sum = 0;
  for (x=0; x<sqr(2*RADIUS+1); x++) sum += matrix[x];
  matrix[RADIUS*(2*RADIUS+1)+RADIUS] += 256 - sum;
  sum = 0;
  for (x=0; x<sqr(2*RADIUS+1); x++) sum += matrix[x];
  DebugLevel(1, "sum=%d\n", sum);

  ConvolveMass256_nonexpanding(m, matrix, RADIUS);
}

void DoOldFlattening(Mass * m)
{
  if (m->flatteningMovement) {
    int movex, movey;
    Point v;
    movex = 0; movey = 0;
    v = AngleToVector(m->heading, MIN(m->velocity, 256));
    if (abs(v.x) > randarea(0, 256)) movex = sign(v.x);
    if (abs(v.y) > randarea(0, 256)) movey = sign(v.y);

    /* FIXME: remove those parameters from the mass, they belong only
       to this algo */
    m->flatteningSHR[LEFT] = m->normalSHR;
    m->flatteningSHR[RIGHT] = m->normalSHR;
    m->flatteningSHR[UP] = m->normalSHR;
    m->flatteningSHR[DOWN] = m->normalSHR;
    /* flattening algo SHR movement */
    switch (movex) {
    case -1:
      m->flatteningSHR[LEFT] = m->strongSHR;
      m->flatteningSHR[RIGHT] = m->weakSHR;
      break;
    case +1:
      m->flatteningSHR[LEFT] = m->weakSHR;
      m->flatteningSHR[RIGHT] = m->strongSHR;
      break;
    default:
      m->flatteningSHR[LEFT] = m->normalSHR;
      m->flatteningSHR[RIGHT] = m->normalSHR;
    }
    switch (movey) {
    case -1:
      m->flatteningSHR[UP] = m->strongSHR;
      m->flatteningSHR[DOWN] = m->weakSHR;
      break;
    case +1:
      m->flatteningSHR[UP] = m->weakSHR;
      m->flatteningSHR[DOWN] = m->strongSHR;
      break;
    default:
      m->flatteningSHR[UP] = m->normalSHR;
      m->flatteningSHR[DOWN] = m->normalSHR;
    }
    //DebugLevel(0, "Heading=%d, movex=%d, movey=%d\n", m->heading, movex, movey);
  }

  /* the evil big old flattening algo - but simply the best ;-) */
  { 
    int x, y;
    int flatteningValueR, flatteningValueL, oldHeight;
    FieldPoint * p;
    MassInterface * mi;
    /* flatten horizontally.... */
    for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
      for (x = m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
	p = FieldXY(x, y);
	if (p->owner == m) {
	  /* special threatement of the sector start point */
	  oldHeight = p->height;
	  mi = FindMassInterface(m, FieldXY(x-1, y)->owner);
	  flatteningValueR = (FieldXY(x-1, y)->height * mi->borderReaction) >> (m->flatteningSHR[RIGHT] + 8);
	  /* give something to the right neighbour */
	  m->bufferMass -= flatteningValueR;
	  p->height += flatteningValueR;
	  /* calculate flatteningValue for the new point */
	  flatteningValueL = oldHeight >> m->flatteningSHR[LEFT];
	  flatteningValueR = oldHeight >> m->flatteningSHR[RIGHT];
	  /* give something to the left neighbour */
	  p->height -= flatteningValueL;
	  m->bufferMass += flatteningValueL;
	  /* necessary because mi->borderContact may be negative */
	  if (p->height < flatteningValueR) {
	    m->bufferMass += p->height - flatteningValueR;
	    p->height = flatteningValueR;
	  }
	  /* normal threatement of the center points */
	  while (FieldXY(x+1,y)->owner == m) {
	    oldHeight = FieldXY(x+1, y)->height;
	    /* give something to the right neighbour */
	    p->height -= flatteningValueR;
	    FieldXY(x+1,y)->height += flatteningValueR;
	    x++;
	    p = FieldXY(x, y);
	    /* calculate flatteningValue for the new point */
	    flatteningValueR = oldHeight >> m->flatteningSHR[RIGHT];
	    flatteningValueL = oldHeight >> m->flatteningSHR[LEFT];
	    /* give something to the left neighbour */
	    p->height -= flatteningValueL;
	    FieldXY(x-1,y)->height += flatteningValueL;
	  }
	  /* special threatement of the sector end point */
	  mi = FindMassInterface(m, FieldXY(x+1, y)->owner);
	  flatteningValueL = (FieldXY(x+1, y)->height * mi->borderReaction) >> (m->flatteningSHR[LEFT] + 8);
	  flatteningValueL -= flatteningValueR;
	  p->height += flatteningValueL;
	  m->bufferMass -= flatteningValueL;
	  /* necessary because mi->borderContact may be negative */
	  if (p->height < 0) {
	    m->bufferMass += p->height;
	    p->height = 0;
	  }
	  x++;
	  /* the current position is known to be a border pixel outside
	     the mass. The next pixel to check ist x+1, the for loop
	     will increment x once more. */
	}
      }
    }
    /* ... and (CODE DUPLICATION - KEEP IN SYNC!) vertically. */
    /* comments removed. */
    for (x = m->maxRect.x; x<m->maxRect.x+m->maxRect.w; x++) {
      for (y=m->maxRect.y; y<m->maxRect.y+m->maxRect.h; y++) {
	p = FieldXY(x, y);
	if (p->owner == m) {
	  oldHeight = p->height;
	  mi = FindMassInterface(m, FieldXY(x, y-1)->owner);
	  flatteningValueR = (FieldXY(x, y-1)->height * mi->borderReaction) >> (m->flatteningSHR[DOWN] + 8);
	  m->bufferMass -= flatteningValueR;
	  p->height += flatteningValueR;
	  flatteningValueL = oldHeight >> m->flatteningSHR[UP];
	  flatteningValueR = oldHeight >> m->flatteningSHR[DOWN];
	  p->height -= flatteningValueL;
	  m->bufferMass += flatteningValueL;
	  if (p->height < flatteningValueR) {
	    m->bufferMass += p->height - flatteningValueR;
	    p->height = flatteningValueR;
	  }
	  while (FieldXY(x,y+1)->owner == m) {
	    oldHeight = FieldXY(x, y+1)->height;
	    p->height -= flatteningValueR;
	    FieldXY(x,y+1)->height += flatteningValueR;
	    y++;
	    p = FieldXY(x, y);
	    flatteningValueR = oldHeight >> m->flatteningSHR[DOWN];
	    flatteningValueL = oldHeight >> m->flatteningSHR[UP];
	    p->height -= flatteningValueL;
	    FieldXY(x,y-1)->height += flatteningValueL;
	  }
	  mi = FindMassInterface(m, FieldXY(x, y+1)->owner);
	  flatteningValueL = (FieldXY(x, y+1)->height * mi->borderReaction) >> (m->flatteningSHR[UP] + 8);
	  flatteningValueL -= flatteningValueR;
	  p->height += flatteningValueL;
	  m->bufferMass -= flatteningValueL;
	  if (p->height < 0) {
	    m->bufferMass += p->height;
	    p->height = 0;
	  }
	  y++;
	}
      }
    }
  }
}

void DoFlattening(Mass * m)
{
  DistributeBufferMass(m);
  if (localPlayer[0].button2) {
    DoConvolveFlattening(m);
  } else {
    DoOldFlattening(m);
  }
  /* do the flattening with the convolution matrix algo */
  //unsigned convolveFlattening : 1;
  /* do the flattening with the old (fast but weak) algo */
  //unsigned oldFlattening : 1;
}

int ContinueMass(Mass * m)
{
  ReThink(m);
  PreFlattening(m);
#ifdef HEAVY_DEBUG
  DebugCheckField();
#endif
  /* the functions below must not delete/create any masses */
  UpdateInterfaces(m);
  RenewInterfaceReactions(m);
  HandleBorderPoints(m);
#ifdef HEAVY_DEBUG
  //DebugCheckField();
#endif
  DoFlattening(m);

#ifdef HEAVY_DEBUG
  DebugCheckField();
  DebugCheckMass(m);
  DebugLevel(3, "Post-Flattening:\n");
  DebugPrintMass(m);
#endif

  if (m->numPixels == 0) {
    DestroyMass(m);
    return 1;
  }
  return 0;
}

