#include "masses.h"
#include "mass.h"
#include "field.h"
#include "networkplayer.h"
#include "debug.h"

#include <stdlib.h>
#include <math.h>

Mass * CreateProjectilMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  m->isProjectil = 1;
  m->vivid = 1;
  RGBSET(m->baseColor, 255, 155, 111);
  m->speed = 256;
  m->growRounds = 90;
  DebugLevel(2, "Created Projectil at (%d,%d)\n", x, y);
  return m;
}

Mass * CreateBombMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  m->isBomb = 1;
  RGBSET(m->baseColor, 255, 255, 255);
  m->speed = 256;
  m->growRounds = 140;
  DebugLevel(2, "Created Bomb at (%d,%d)\n", x, y);
  return m;
}

Mass * CreateGrowingFoodMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  m->isPlant = 1;
  RGBSET(m->baseColor, 120, 255, 100);
  m->speed = 256;
  DebugLevel(2, "Created GrowinFoodMass at (%d,%d)\n", x, y);
  return m;
}

Mass * CreatePlant2Mass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  m->isPlant2 = 1;
  m->growRounds = 10;
  RGBSET(m->baseColor, 140, 255, 120);
  m->speed = 256;
  m->vivid2 = 0;
  m->minBorderHeight = HEIGHT_NORM;
  //m->maxBorderHeight /= 2;
#if 0
  m->baseLineGrowth = calloc(1, sizeof(BaseLineGrowth));
  m->baseLineGrowth->step = 128;
  m->baseLineGrowth->height = 0;
  m->baseLineGrowth->start = 0;
  m->baseLineGrowth->end = 0;	/* will be set later */
#endif
  DebugLevel(2, "Created Plant2Mass at (%d,%d)\n", x, y);
  return m;
}

Mass * CreateLivingWallMass(int x, int y)
{
  Mass * m;
  m = CreateGrowingWallMass(x, y);
  if (!m) return NULL;
  RGBSET(m->baseColor, 230, 220, 220);
  m->vivid = 0;
  m->vivid2 = 2;
  DebugLevel(2, "Turned to LivingWallMass (%d,%d)\n", x, y);
  return m;
}

Mass * CreateGrowingWallMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  RGBSET(m->baseColor, 120, 120, 120);
  m->isWall = 1;
  /* no spike! */
  /* m->bufferMass += HEIGHT_NORM*4000; */
  m->growRounds = 90;
  m->superGrow = 1;
  m->speed = 10*256;
  //m->minBorderHeight *= 2;
  //m->maxBorderHeight *= 2;
  DebugLevel(2, "Created GrowingWallMass at (%d,%d)\n", x, y);
  return m;
}  

Mass * CreateAvoidPointMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  RGBSET(m->baseColor, 120, 220, 120);
  m->isWall = 0;
  m->growRounds = randarea(700, 3000);
  m->speed = 256;
  m->vivid2 = 4;
  DebugLevel(2, "Created AvoidPointMass at (%d,%d)\n", x, y);
  return m;
}  

/*
** Player mass
*/

Mass * CreatePlayerMass(int x, int y, int playernr)
{
  Mass * m;
  m = CreateMassNearby(x, y, 10);
  m->player = playernr;
  m->growRounds = 940;
  m->flatteningMovement = 1;
  
  switch (m->player) {
  case 0:
    RGBSET(m->baseColor, 0, 255, 255);
    /* important: set the screen center to the player, else the player
       might be in the frozen area and never set the screen center */
    fieldCenter.x = x;
    fieldCenter.y = y;
    break;
  case 1:
    RGBSET(m->baseColor, 255, 0, 255);
    break;
  default:
    assert(0);
  }
  DebugLevel(2, "Created PlayerMass for player %d\n", m->player);
  return m;
}

Mass * CreateMazeWall(int x, int y)
{
  Mass * m;
  m = CreateGrowingWallMass(x, y);
  m->growRounds = 1000;
  m->isMazeWall = 1;
  return m;
}

Mass * CreateMassattackMass(Point target)
{
  int x, y, i;
  Point v;
  Mass * m;
  for (i=100; i>0; i--) {
    x = randarea(0, FIELDSIZE);
    y = randarea(0, FIELDSIZE);
    if ((m = CreateMassNearby(x, y, 1))) break;
  }
  if (!m) return NULL;
  RGBSET(m->baseColor, 255, 20, 20);
  m->isMeltingMass = 1;
  m->growRounds = 250;
  m->speed = 256;
  v.x = target.x-x;
  v.y = target.y-y;
  m->heading = VectorToAngle(v);
  m->flatteningMovement = 1;
  //m->weakSHR--;
  //m->normalSHR--;
  //m->strongSHR--;
  DebugLevel(2, "Created MassattackMass\n");
  return m;
}

Mass * CreateCoinMass(int px, int py)
{
  Mass * m;
  int i;
  m = CreateMassNearby(px, py, 1);
  if (!m) return NULL;
  for (i=0; i<40; i++) AddBorderPixel(m);
  m->oldFlattening = 0;
  m->speed = 256;
  m->isCoinMass = 1;
  m->displayType = DT_FLAT;
  RGBSET(m->baseColor, 255, 255, 80);
  return m;
}

//static char maze[10][10];
void MazeGenerator()
{
  /* process all maze masses */
  //int i, j;
  /*
  for (i=0; i<numMasses; i++) {
    for (j=i+1; j<numMasses; j++) {
      if (masses[i]->isMazeWall && masses[j]->isMazeWall) {
	if (randarea(0, 20) == 0) {
	  MassInterface * mi;
	  mi = CreateInterface(masses[i], masses[j]);
	}
      }
    }
  }
  */
  /* OUTDATED:
  for (i=1; i<numMasses; i++) {
    j = i-1;
    if (masses[i]->isMazeWall && masses[j]->isMazeWall) {
      MassInterface * mi;
      mi = CreateInterface(masses[i], masses[j]);
    }
  }
  */
}

Mass * CreateCellMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  if (!m) return NULL;
  m->cell = calloc(1, sizeof(CellInfo)); 
  m->cell->golState = randarea(0, 2);
  m->cell->wishDir = randarea(0, 256);
  m->cell->pulse = 1;
  m->velocity = 0;
  m->flatteningMovement = 1;
  m->growRounds = 350;
  //m->vivid2 = 1;
  RGBSET(m->baseColor, 100, 100, 55);
  m->speed = 256;
  DebugLevel(2, "Created CellMass at (%d,%d)\n", x, y);
  return m;
}

Mass * CreateNexusMass(int x, int y)
{
  Mass * m;
  m = CreateMassNearby(x, y, 1);
  /* the maximum! */
  m->nexusLevel = 5;
  m->lookDistance = 40;
  RGBSET(m->baseColor, 255, 255, 255);
  return m;
}
