#include <stdio.h>
#include <stdlib.h>
#include "math2d.h"
#include "debug.h"

void ExpandRectToIncludePoint(Rect * r, int x, int y) 
{
  if (r->w == 0) {
    r->w = 1; r->h = 1;
    r->x = x; r->y = y;
  } else {
    if (x < r->x) { r->w += r->x-x; r->x = x; } else
    if (x >= r->x+r->w) { r->w = x - r->x + 1; }

    if (y < r->y) { r->h += r->y-y; r->y = y; } else
    if (y >= r->y+r->h) { r->h = y - r->y + 1; }
  }
  DebugLevel(4, "Included Point (%d, %d)\n", x, y);
}

Rect SmallestCommonRect(Rect r1, Rect r2) {
  /* FIXME: quickly check this code for correctness. */
  Rect r;
  assert(r1.h>=0 && r1.w>=0);
  assert(r2.h>=0 && r2.w>=0);
  r.x = MAX(r1.x, r2.x);
  r.y = MAX(r1.y, r2.y);
  r.w = MIN(r1.x+r1.w, r2.x+r2.w)-r.x;
  r.h = MIN(r1.y+r1.h, r2.y+r2.h)-r.y;
  /* this is a valid case */
  if (r.h < 0) r.h = 0;
  if (r.w < 0) r.w = 0;
  return r;
}

double GetRealDistance(Point p1, Point p2)
{
  return sqrt(
    sqr((double)p1.x-(double)p2.x) + 
    sqr((double)p1.y-(double)p2.y)
  );
}

int GetCompareDistance(Point p1, Point p2)
{
  return sqr(p1.x-p2.x) + sqr(p1.y-p2.y);
}

double GetRealLength(Point v)
{
  return sqrt(sqr((double)v.x) + sqr((double)v.y) );
}

int GetCompareLength(Point v)
{
  return sqr(v.x) + sqr(v.y);
}

void FitPointIntoRect(Point * p, Rect r)
{
  if (p->x < r.x) {
    p->x = r.x;
  } else if (p->x > r.x+r.w-1) {
    p->x = r.x+r.w-1;
  }
  if (p->y < r.y) {
    p->y = r.y;
  } else if (p->y > r.y+r.h-1) {
    p->y = r.y+r.h-1;
  }
}

Point AngleToVector(int angle, int length)
{
  /* FIXME: don't use double: must run synchronous across platforms */
  Point res;
  /* so no adaption is necessary in OpenGL code */
  angle -= 256/4;
  res.x = (double)length * cos((double)angle/256 * 2*M_PI);
  res.y = (double)length * sin((double)angle/256 * 2*M_PI);
  return res;
}

int VectorToAngle(Point v)
{
  /* FIXME: don't use double: must run synchronous across platforms */
  if (v.x == 0 && v.y == 0) return 0;
  /* so no adaption is necessary in OpenGL code */
  return (int)(atan2(v.y, v.x)/(M_PI*2)*256) + 256/4;
}

Point VectorFromTo(Point from, Point to)
{
  Point res;
  res.x = to.x - from.x;
  res.y = to.y - from.y;
  return res;
}

Point AddVectors(Point v1, Point v2)
{
  Point res;
  res.x = v1.x + v2.x;
  res.y = v1.y + v2.y;
  return res;
}

Point SubVectors(Point v1, Point v2)
{
  Point res;
  res.x = v1.x - v2.x;
  res.y = v1.y - v2.y;
  return res;
}

Point XYToPoint(int x, int y)
{
  Point res;
  res.x = x;
  res.y = y;
  return res;
}

char * RectString(Rect r)
{
  char * s = malloc(255);
  sprintf(s, "{x=%d, y=%d, w=%d, h=%d}", r.x, r.y, r.w, r.h);
  return s;
}

char * PointString(Point p)
{
  char * s = malloc(255);
  sprintf(s, "{x=%d, y=%d}", p.x, p.y);
  return s;
}
