#ifndef __MATH2D__
#define __MATH2D__

#include <math.h>

/* Some <math.h> files do not define M_PI... */
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef __sgi
/* Most math.h's do not define float versions of the trig functions. */
#define sinf sin
#define cosf cos
#define atan2f atan2
#endif

/* the OGL part uses those */
#define RAD(x) (((x)*M_PI)/180.0)
#define DEG(x) (((x)*180.0)/M_PI)

/* note: integer, not char, to make calculations easier (range: 0-255) */
typedef int Color[3];
#define RED 0
#define GREEN 1
#define BLUE 2
#define RGBSET(c,r,g,b) c[0]=r; c[1]=g; c[2]=b;

typedef enum { LEFT=0, RIGHT=1, UP=2, DOWN=3 } Direction;

typedef struct { int x, y, w, h; } Rect;
typedef struct { int x, y; } Point;

#define sqr(x) ((x)*(x))
#define MIN(a, b) ((a)<(b)?(a):(b))
#define MAX(a, b) ((a)>(b)?(a):(b))
#define sign(x) ((x)==0?0:((x)>0?1:-1))

/*  */
extern void ExpandRectToIncludePoint(Rect * r, int x, int y);
extern Rect SmallestCommonRect(Rect r1, Rect r2);
#define IsPointInRect(p, r) (p.x >= r.x && p.y >= r.y && p.x < r.x+r.w && p.y < r.y+r.h && r.w>0)
extern double GetRealDistance(Point p1, Point p2);
extern int GetCompareDistance(Point p1, Point p2);

extern double GetRealLength(Point v);
extern int GetCompareLength(Point v);

extern void FitPointIntoRect(Point * p, Rect r);

extern Point VectorFromTo(Point from, Point to);
extern Point AddVectors(Point v1, Point v2);
extern Point SubVectors(Point v1, Point v2);

Point XYToPoint(int x, int y);

/* angles are measured from 0..255, with 256 = 0 */
extern Point AngleToVector(int angle, int length);
extern int VectorToAngle(Point v);

/* for debugging; you should free the memory, but who cares... */
extern char * RectString(Rect r);
extern char * PointString(Point p);

#endif
