#include <stdio.h>
#include <string.h>
#include "field.h"

int tests_failed=0;
int tests_done=0;

#define test(cond, comment) do_test(cond, comment, __FILE__,  __LINE__)
int do_test(int cond, char * comment, char * file, int line)
{
  tests_done++;
  if (cond) {
    fprintf(stderr, "OK: %s\n", comment);
  } else {
    fprintf(stderr, "%s:%d: FAILED: %s\n", file, line, comment);
    tests_failed++;
  }
  return cond;
}

void Test_Field()
{
  Rect r1, r2, exp, res;
  Point p1, p2, pres;
  int i, j;

  i = 3;             j = 12;
  i += 20*FIELDSIZE; j -= 7*FIELDSIZE;
  i &= FIELDWRAP;    j &= FIELDWRAP;
  test(i==3 && j == 12, "Wrap logic check");

  r1.x = 0;  r1.y = 30;
  r1.w = 20; r1.h = 20;

  r2.x = FIELDSIZE-5; r2.y = FIELDSIZE-5;
  r2.w = 20;          r2.h = 70;

  r1.x -= 2*FIELDSIZE;
  r1.y += 5*FIELDSIZE;
  res = Wrap_SmallestCommonRect(r1, r2);
  res.x &= FIELDWRAP;
  res.y &= FIELDWRAP;
  exp.x = 0;  exp.y = 30;
  exp.w = 15; exp.h = 20;
  test(memcmp(&res, &exp, sizeof(Rect)) == 0, "Wrap_SmallestCommonRect");

  r1.x = 0;  r1.y = 30;
  r1.w = 20; r1.h = 20;
  p1.x = FIELDSIZE-3; p1.y = r1.y + 5;
  pres.x = 0;
  pres.y = r1.y + 5;
  p1.x -= 2*FIELDSIZE;
  r1.x -= 2*FIELDSIZE;
  r1.y += 5*FIELDSIZE;
  Wrap_FitPointIntoRect(&p1, r1);
  p1.x &= FIELDWRAP;
  p1.y &= FIELDWRAP;
  test(memcmp(&p1, &pres, sizeof(Point)) == 0, "Wrap_FitPointIntoRect");
  
  p1.x = 2;  p1.y = 3;
  p2.x = -2; p2.y = 3;
  i = Wrap_GetCompareDistance(p1, p2);
  p1.y -= FIELDSIZE;
  p2.y += 4*FIELDSIZE;
  j = Wrap_GetCompareDistance(p1, p2);
  test(i==j, "Wrap_GetCompareDistance 1");
  p2.y -= 1;
  j = Wrap_GetCompareDistance(p1, p2);
  test(j>i, "Wrap_GetCompareDistance 2");
}

int main()
{
  /* stderr because mass debug messages go to stdout */
  fprintf(stderr, "Doing some self-checks.\n");

  Test_Field();

  /* summary */
  if (!tests_failed) {
    fprintf(stderr, "All %d tests succeeded.\n", tests_done);
  } else {
    fprintf(stderr, "%d tests of %d failed.\n", tests_failed, tests_done);
  }
  exit(tests_failed);
}
