#include "video3d.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "field.h"
#include "debug.h"
#include "networkplayer.h"
#include "input.h"
#include "math2d.h"
#include "game.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL.h>

#ifdef OGL

#define DISPLAY_BPP 0 //default
/* TODO: get this from video2d.h */
#define DISPLAY_SCREENX 256
#define DISPLAY_SCREENY 256

/* the displayed screen rectangle, in mass coordinates */
Rect OGLscreenRect = {0,0,DISPLAY_SCREENX,DISPLAY_SCREENY};

#define WINDOW_SIZEX 640
#define WINDOW_SIZEY 640

void OGLInit(int width, int height) {
  glViewport(0, 0, WINDOW_SIZEX, WINDOW_SIZEY);
  glClearColor(0., 0., 0., 1.);
  glClearDepth(1.0);
  glLineWidth(1.);				
  glDepthFunc(GL_LESS);				
  glEnable(GL_DEPTH_TEST);
  glShadeModel(GL_SMOOTH);

//Back face culling 
  glFrontFace(GL_CCW);
  glEnable(GL_CULL_FACE);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
}

/* 3D and 2D at the same time */
int InitVideo3d(int argc, char ** argv) { 
 
  /* Initialize SDL for video output */
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
    return 0;
  }

  /* Create a 640x480 OpenGL screen */
  if ( SDL_SetVideoMode(WINDOW_SIZEX, WINDOW_SIZEY, 0, SDL_OPENGL) == NULL ) {
    fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
    SDL_Quit();
    return 0;;
  }

  /* Set the title bar in environments that support it */
  SDL_WM_SetCaption("MASS", NULL);

  /* Loop, drawing and checking events */
  OGLInit(WINDOW_SIZEX, WINDOW_SIZEY);

  //  DebugLevel(1, "SDL 3D frontend ready.\n");
  return 0;
}

float getFPS(int count) {
  static clock_t told=0, tnew;
  static int countold=0;
  int numcount;

  told = tnew;
  tnew = clock();
  numcount = count-countold;
  countold = count;
  return ((float)(numcount)/((float)(tnew-told)/CLOCKS_PER_SEC));
}

void showFPS(void) {
  static int count=0;
  static char frbuf[20], fps[30];
  int len;
/*
  count++;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-0.5, 639.5, -0.5, 479.5, -1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glDisable(GL_LIGHTING);

  glColor3f(1., 1., 0.);

  count++;
  if (!(count % 50))
    sprintf(fps,"FPS: %f", getFPS());
  glRasterPos2i(10, 460);
  len=(int)strlen(fps);
  for(i=0;i<len;i++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, fps[i]);

  glRasterPos2i(10, 440);
  sprintf(frbuf,"%d Masses", numMasses);
  len=(int)strlen(frbuf);
  for(i=0;i<len;i++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, frbuf[i]);
*/
}

void OGL_Draw_2D_Mass(Mass * m) {
  int x, y;
  FieldPoint * p;
  Rect visible;	
  visible = Wrap_SmallestCommonRect(OGLscreenRect, m->maxRect);
  for (y=visible.y; y<visible.y+visible.h; y++) {
    for (x=visible.x; x<visible.x+visible.w; x++) {
      p = FieldXY(x, y);
      if (p->owner == m) {
	glColor3ub(m->baseColor[0], m->baseColor[1], m->baseColor[2]);
	glVertex2f((x-OGLscreenRect.x)&FIELDWRAP, DISPLAY_SCREENY-(y-OGLscreenRect.y)&FIELDWRAP);
      }
    }
  }
}

void OGL_Draw_2D (Mass ** visibleMasses) {
  int i;

  glViewport(WINDOW_SIZEX - DISPLAY_SCREENX, WINDOW_SIZEY - DISPLAY_SCREENY,
	     DISPLAY_SCREENX, DISPLAY_SCREENY);
  //set up the Projection Matrix
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, DISPLAY_SCREENX, 0, DISPLAY_SCREENY, 1, -1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glDisable(GL_LIGHTING);

  glBegin(GL_POINTS);
    for (i=0; visibleMasses[i]; i++) OGL_Draw_2D_Mass(visibleMasses[i]);
  glEnd();
}

void OGL_Draw_3D_Mass(Mass * m) {
  int x, y, displayx, displayy;
  /* TODO: common code with DrawMass, OGL_Draw_2D_Mass*/
  FieldPoint * p;
  Rect visible;	 	
  /* This assumes (FIELDSIZE > MAX(SCREENX,SCREENY)). No assertion
     here because a small FIELDSIZE is useful for debugging. */
  /* set "visible" to the screen rect, expressed in mass coordinates */
  visible = Wrap_SmallestCommonRect(OGLscreenRect, m->maxRect);
  for (y=visible.y-1; y<visible.y+visible.h+1; y+=1) {
//for not wireframe rendering use GL_TRIANGLE_STRIP
     glBegin(GL_LINE_STRIP);
    for (x=visible.x-1; x<visible.x+visible.w+1; x+=2) {
      displayx = (x-OGLscreenRect.x)&FIELDWRAP;
      displayy = (y-OGLscreenRect.y)&FIELDWRAP;
      p = FieldXY(x, y);
      if (p->owner) {
	glColor3ub(m->baseColor[0], m->baseColor[1], m->baseColor[2]);
	glVertex3f(displayx, p->height, displayy);
      } else {
	glColor3f(0., 0., 0.);
        glVertex3f(displayx, 0., displayy);
      }
      p = FieldXY(x, y+1);
      if (p->owner) {
	glColor3ub(m->baseColor[0], m->baseColor[1], m->baseColor[2]);
	glVertex3f(displayx, p->height, displayy+1);
      } else {
	glColor3f(0., 0., 0.);
        glVertex3f(displayx, 0., displayy+1);
      }
      p = FieldXY(x+1, y);
      if (p->owner) {
	glColor3ub(m->baseColor[0], m->baseColor[1], m->baseColor[2]);
	glVertex3f(displayx+1, p->height, displayy);
      } else {
	glColor3f(0., 0., 0.);
	glVertex3f(displayx+1, 0., displayy);
      }
      p = FieldXY(x+1, y+1);
      if (p->owner) {
	glColor3ub(m->baseColor[0], m->baseColor[1], m->baseColor[2]);
	glVertex3f(displayx+1, p->height, displayy+1);
      } else {
	glColor3f(0., 0., 0.);
	glVertex3f(displayx+1, 0., displayy+1);
      }
    }
    glEnd();/*GL_TRIANGLE_STRIP*/
  }
}

void OGL_Draw_3D(Mass ** visibleMasses) {
  int i;

  glMatrixMode(GL_PROJECTION);
  glViewport(0, 0, WINDOW_SIZEX, WINDOW_SIZEY);
  glLoadIdentity();
  glFrustum(-.41, .41, -.41, .41, 1, 1000);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

//viewer pos 20 is an arbitrary value
  glTranslatef(0.,-20., 0);
  glRotatef(displayHeading*1.411, 0., 1., 0.);

//center the field
  glTranslatef(-DISPLAY_SCREENX/2, 0., -DISPLAY_SCREENY/2);
//squeeze height to a resonable value
  glScalef(1., 1./(float)HEIGHT_NORM, 1.);

  for (i=0; visibleMasses[i]; i++) OGL_Draw_3D_Mass(visibleMasses[i]);;
}

void DoneVideo3d() { }
void DrawFrame3d() {
  Mass ** visibleMasses;
  //wipe out all remainders 
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* top left corner of the screen in mass coordinates */
  OGLscreenRect.x = displayCenter.x - DISPLAY_SCREENX/2;
  OGLscreenRect.y = displayCenter.y - DISPLAY_SCREENY/2;

//compute visible masses for 2d and 3d
  visibleMasses = GetMassesInRect(OGLscreenRect);

  OGL_Draw_2D(visibleMasses);
  OGL_Draw_3D(visibleMasses);

//free memory
  free(visibleMasses);

  //swap buffers
  SDL_GL_SwapBuffers();
}

#else  /* #ifdef OGL */
int InitVideo3d(int argc, char ** argv) {
  printf("OpenGL not available, the game was compiled without.\n");
  return -1;
}
void DoneVideo3d() { }
void DrawFrame3d() { }
#endif /* #ifdef OGL */
